from __future__ import annotations

from typing import Literal, Optional, Union, List
from xradio.schema.bases import (
    xarray_dataset_schema,
    xarray_dataarray_schema,
    dict_schema,
)
from xradio.schema.typing import Attr, Coord, Coordof, Data, Dataof, Name
import numpy

# Dimensions
Time = Literal["time"]
""" Observation time dimension """
TimeCal = Literal["time_cal"]
""" time dimension of system calibration (when not interpolated to main time)"""
TimeEphemeris = Literal["time_ephemeris"]
""" time dimension of ephemeris data (when not interpolated to main time) """
TimePhaseCal = Literal["time_phase_cal"]
""" Coordinate label for VLBI-specific phase cal time axis """
TimePointing = Literal["time_pointing"]
""" time dimension of pointing dataset (when not interpolated to main time) """
TimeWeather = Literal["time_weather"]
""" time dimension of weather dataset (when not interpolated to main time) """
AntennaName = Literal["antenna_name"]
""" Antenna name dimension """
StationName = Literal["station_name"]
""" Station identifier dimension """
ReceptorLabel = Literal["receptor_label"]
""" Receptor label dimension """
ToneLabel = Literal["tone_label"]
""" Tone label dimension """
BaselineId = Literal["baseline_id"]
""" Baseline ID dimension """
Frequency = Literal["frequency"]
""" Frequency dimension """
FrequencyCal = Literal["frequency_cal"]
""" Frequency dimension in the system calibration dataset """
Polarization = Literal["polarization"]
""" Polarization dimension """
UvwLabel = Literal["uvw_label"]
""" Coordinate dimension of UVW data (typically shape 3 for 'u', 'v', 'w') """
SkyDirLabel = Literal["sky_dir_label"]
""" Coordinate labels of sky directions (typically shape 2 and 'ra', 'dec') """
LocalSkyDirLabel = Literal["local_sky_dir_label"]
""" Coordinate labels of local sky directions (typically shape 2 and 'az', 'alt') """
SphericalDirLabel = Literal["spherical_dir_label"]
""" Coordinate labels of spherical directions (shape 2 and 'lon', 'lat1' """
SkyPosLabel = Literal["sky_pos_label"]
""" Coordinate labels of sky positions (typically shape 3 and 'ra', 'dec', 'dist') """
SphericalPosLabel = Literal["spherical_pos_label"]
""" Coordinate labels of spherical positions (shape shape 3 and 'lon', 'lat1', 'dist2') """
EllipsoidPosLabel = Literal["ellipsoid_pos_label"]
""" Coordinate labels of geodetic earth location data (typically shape 3 and 'lon', 'lat', 'height')"""
CartesianPosLabel = Literal["cartesian_pos_label"]
""" Coordinate labels of geocentric earth location data (typically shape 3 and 'x', 'y', 'z')"""
nPolynomial = Literal["n_polynomial"]
""" For data that is represented as variable in time using Taylor expansion """
PolyTerm = Literal["poly_term"]
""" Polynomial term used in VLBI GAIN_CURVE """
LineLabel = Literal["line_label"]
""" Line labels (for line names and variables). """

# Represents "no dimension", i.e. used for coordinates and data variables with
# zero dimensions.
ZD = tuple[()]


# Types of quantity and measures
Quantity = Literal["quantity"]
SkyCoord = Literal["sky_coord"]
SpectralCoord = Literal["spectral_coord"]
Location = Literal["location"]
Doppler = Literal["doppler"]


# Units of quantities and measures
UnitsSeconds = list[Literal["s"]]
UnitsHertz = list[Literal["Hz"]]
UnitsMeters = list[Literal["m"]]

UnitsOfSkyCoordInRadians = list[Literal["rad"], Literal["rad"]]
UnitsOfLocationInMetersOrRadians = Union[
    list[Literal["m"], Literal["m"], Literal["m"]],
    list[Literal["rad"], Literal["rad"], Literal["m"]],
]
UnitsOfPositionInRadians = list[Literal["rad"], Literal["rad"], Literal["m"]]
UnitsOfDopplerShift = Union[list[Literal["ratio"]], list[Literal["m/s"]]]

UnitsRadians = list[Literal["rad"]]
UnitsKelvin = list[Literal["K"]]
UnitsKelvinPerJansky = list[Literal["K/Jy"]]
UnitsMetersPerSecond = list[Literal["m/s"]]
UnitsPascal = list[Literal["Pa"]]  # hPa? (in MSv2)
UnitsPerSquareMeters = list[Literal["/m^2"]]


# Quantities


@xarray_dataarray_schema
class QuantityInSecondsArray:
    """
    Quantity with units of seconds
    """

    data: Data[ZD, float]

    units: Attr[UnitsSeconds]
    type: Attr[Quantity] = "quantity"


@xarray_dataarray_schema
class QuantityInHertzArray:
    """
    Quantity with units of Hertz
    """

    data: Data[ZD, float]

    units: Attr[UnitsHertz]
    type: Attr[Quantity] = "quantity"


@xarray_dataarray_schema
class QuantityInMetersArray:
    """
    Quantity with units of Hertz
    """

    data: Data[ZD, float]

    units: Attr[UnitsMeters]
    type: Attr[Quantity] = "quantity"


@xarray_dataarray_schema
class QuantityInMetersPerSecondArray:
    """
    Quantity with units of Hertz
    """

    data: Data[ZD, float]

    units: Attr[UnitsMetersPerSecond]
    type: Attr[Quantity] = "quantity"


@xarray_dataarray_schema
class QuantityInRadiansArray:
    """
    Quantity with units of Hertz
    """

    data: Data[ZD, float]

    units: Attr[UnitsRadians]
    type: Attr[Quantity] = "quantity"


@xarray_dataarray_schema
class QuantityInKelvinArray:
    """
    Quantity with units of Kelvins
    """

    data: Data[ZD, float]

    units: Attr[UnitsKelvin]
    type: Attr[Quantity] = "quantity"


@xarray_dataarray_schema
class QuantityInKelvinPerJanskyArray:
    """
    Quantity with units of K/Jy (sensitivity in gain curve)
    """

    data: Data[ZD, numpy.float64]

    units: Attr[UnitsKelvinPerJansky]
    type: Attr[Quantity] = "quantity"


@xarray_dataarray_schema
class QuantityInPascalArray:
    """
    Quantity with units of Pa
    """

    data: Data[ZD, numpy.float64]

    units: Attr[UnitsPascal]
    type: Attr[Quantity] = "quantity"


@xarray_dataarray_schema
class QuantityInPerSquareMetersArray:
    """
    Quantity with units of /m^2
    """

    data: Data[ZD, numpy.float64]

    units: Attr[UnitsPerSquareMeters]
    type: Attr[Quantity] = "quantity"


AllowedTimeScales = Literal["tai", "tcb", "tcg", "tdb", "tt", "ut1", "utc"]


AllowedTimeFormats = Literal["unix", "mjd", "cxcsec", "gps"]


@xarray_dataarray_schema
class TimeArray:
    """
    Representation of a time measure.

    :py:class:`astropy.time.Time` serves as the reference implementation.
    Data can be converted as follows::

        astropy.time.Time(data * astropy.units.Unit(attrs['units'][0]),
                          format=attrs['format'], scale=attrs['scale'])

    All formats that express time as floating point numbers since an epoch
    are permissible, so at present the realistic options are:

    * ``mjd`` (from 1858-11-17 00:00:00 UTC)
    * ``unix`` (from 1970-01-01 00:00:00 UTC)
    * ``unix_tai`` (from 1970-01-01 00:00:00 TAI)
    * ``cxcsec`` (from 1998-01-01 00:00:00 TT)
    * ``gps`` (from 1980-01-06 00:00:00 UTC)

    """

    data: Data[ZD, float]
    """Time since epoch, typically in seconds (see ``units``)."""

    type: Attr[Time] = "time"
    """ Array type. Should be ``"time"``. """
    units: Attr[UnitsSeconds] = ("s",)
    """ Units to associate with axis"""
    scale: Attr[AllowedTimeScales] = "utc"
    """
    Time scale of data. Must be one of ``(‘tai’, ‘tcb’, ‘tcg’, ‘tdb’, ‘tt’, ‘ut1’, ‘utc’)``,
    see :py:class:`astropy.time.Time`
    """
    format: Attr[AllowedTimeFormats] = "unix"
    """Time representation and epoch, see :py:class:`TimeArray`."""


# Taken from the list of astropy built-in frame classes: https://docs.astropy.org/en/stable/coordinates/index.html
AllowedSkyCoordFrames = Literal[
    "icrs",
    "fk5",
    "fk4",
    "fk4noterms",
    "galactic",
    "galactocentric",
    "supergalactic",
    "altaz",
    "hadec",
    "gcrs",
    "cirs",
    "itrs",
    "hcrs",
    "teme",
    "tete",
    "precessedgeocentric",
    "geocentricmeanecliptic",
    "barycentricmeanecliptic",
    "heliocentricmeanecliptic",
    "geocentrictrueecliptic",
    "barycentrictrueecliptic",
    "heliocentrictrueecliptic",
    "heliocentriceclipticiau76",
    "custombarycentricecliptic",
    "lsr",
    "lsrk",
    "lsrd",
    "galacticlsr",
]


@xarray_dataarray_schema
class SkyCoordArray:
    """Measures array for data variables that are sky coordinates, used in :py:class:`FieldSourceXds`"""

    data: Data[Union[SkyDirLabel, SkyPosLabel], float]

    type: Attr[SkyCoord] = "sky_coord"
    units: Attr[UnitsOfSkyCoordInRadians] = ("rad", "rad")
    frame: Attr[AllowedSkyCoordFrames] = ""
    """
    Possible values are astropy SkyCoord frames.
    Several casacore frames found in MSv2 are translated to astropy frames as follows: AZELGEO=>altaz, J2000=>fk5, ICRS=>icrs.
    From fixvis docs: clean and the im tool ignore the reference frame
    claimed by the UVW column (it is often mislabelled as ITRF when it is
    really FK5 (J2000)) and instead assume the (u, v, w)s are in the same frame
    as the phase tracking center. calcuvw does not yet force the UVW column and
    field centers to use the same reference frame! Blank = use the phase
    tracking frame of vis.
    """


@xarray_dataarray_schema
class LocalSkyCoordArray:
    """Measures array for the arrays that have coordinate local_sky_dir_label in :py:class:`PointingXds`"""

    data: Data[LocalSkyDirLabel, float]

    type: Attr[SkyCoord] = "sky_coord"
    units: Attr[UnitsOfSkyCoordInRadians] = ("rad", "rad")
    frame: Attr[AllowedSkyCoordFrames] = "fk5"
    """
    From fixvis docs: clean and the im tool ignore the reference frame claimed by the UVW column (it is often mislabelled
    as ITRF when it is really FK5 (J2000)) and instead assume the (u, v, w)s are in the same frame as the phase tracking
    center. calcuvw does not yet force the UVW column and field centers to use the same reference frame! Blank = use the
    phase tracking frame of vis.
    """


# Coordinates / Axes
@xarray_dataarray_schema
class TimeCoordArray:
    """Data model of the main dataset time axis. See also :py:class:`TimeArray`."""

    data: Data[Time, float]
    """
    Time, expressed in seconds since the epoch (see ``scale`` &
    ``format``), see also see :py:class:`TimeArray`.
    """

    type: Attr[Time] = "time"
    """ Coordinate type. Should be ``"time"``. """

    units: Attr[UnitsSeconds] = ("s",)
    """ Units to associate with axis"""

    scale: Attr[AllowedTimeScales] = "utc"
    """ Astropy time scales, see :py:class:`TimeArray` """

    format: Attr[AllowedTimeFormats] = "unix"
    """ Astropy format, see :py:class:`TimeArray`"""

    integration_time: Attr[QuantityInSecondsArray] = None
    """ The nominal sampling interval (ms v2). Units of seconds. """


@xarray_dataarray_schema
class TimeInterpolatedCoordArray:
    """
    Data model of a time axis when it is interpolated to match the time
    axis of the main dataset. This can be used in the system_calibration_xds,
    pointing_xds, weather_xds, field_and_source_info_xds, and phase_cal_xds
    when their respective time_cal, time_pointing, time_weather,
    time_ephemeris or time_phase_cal are interpolated to the main dataset
    time. See also :py:class:`TimeArray`.

    The only difference with respect to the main TimeCoordArray is the
    absence of the attribute integration_time
    """

    data: Data[Time, float]
    """
    Time, expressed in seconds since the epoch (see ``scale`` &
    ``format``), see also see :py:class:`TimeArray`.
    """

    type: Attr[Time] = "time"
    """ Coordinate type. Should be ``"time"``. """

    units: Attr[UnitsSeconds] = ("s",)
    """ Units to associate with axis"""

    scale: Attr[AllowedTimeScales] = "utc"
    """ Astropy time scales, see :py:class:`TimeArray` """

    format: Attr[AllowedTimeFormats] = "unix"
    """ Astropy format, see :py:class:`TimeArray`"""


@xarray_dataarray_schema
class TimeCalCoordArray:
    """Data model of 'time_cal' axis (time axis in system_calibration_xds
    subdataset when not interpolated to the main time axis. See also
    :py:class:`TimeCoordArray`."""

    data: Data[TimeCal, float]
    """
    Time, expressed in seconds since the epoch (see ``scale`` &
    ``format``).
    """

    type: Attr[Time] = "time_cal"
    """ Coordinate type. Should be ``"time_cal"``. """

    units: Attr[UnitsSeconds] = ("s",)
    """ Units to associate with axis"""

    scale: Attr[AllowedTimeScales] = "utc"
    """ Astropy time scales, see :py:class:`TimeArray` """

    format: Attr[AllowedTimeFormats] = "unix"
    """ Astropy format, see :py:class:`TimeArray`"""


@xarray_dataarray_schema
class TimePointingCoordArray:
    """Data model of the 'time_pointing' axis (time axis in pointing_xds
    subdataset when not interpolated to the main time axis. See also
    :py:class:`TimeCoordArray`."""

    data: Data[TimePointing, float]
    """
    Time, expressed in seconds since the epoch (see ``scale`` &
    ``format``).
    """

    type: Attr[TimePointing] = "time_pointing"
    """ Coordinate type. Should be ``"time_pointing"``. """

    units: Attr[UnitsSeconds] = ("s",)
    """ Units to associate with axis"""

    scale: Attr[AllowedTimeScales] = "utc"
    """ Astropy time scales, see :py:class:`TimeArray` """

    format: Attr[AllowedTimeFormats] = "unix"
    """ Astropy format, see :py:class:`TimeArray`"""


@xarray_dataarray_schema
class TimeEphemerisCoordArray:
    """Data model of the 'time_ephemeris' axis (time axis in the
    field_and_source_info_xds subdataset when not interpolated to the main
    time axis. See also :py:class:`TimeCoordArray`."""

    data: Data[TimeEphemeris, float]
    """
    Time, expressed in seconds since the epoch (see ``scale`` &
    ``format``).
    """

    type: Attr[TimeEphemeris] = "time_ephemeris"
    """ Coordinate type. Should be ``"time_ephemeris"``. """

    units: Attr[UnitsSeconds] = ("s",)
    """ Units to associate with axis"""

    scale: Attr[AllowedTimeScales] = "utc"
    """ Astropy time scales, see :py:class:`TimeArray` """

    format: Attr[AllowedTimeFormats] = "unix"
    """ Astropy format, see :py:class:`TimeArray`"""


@xarray_dataarray_schema
class TimeWeatherCoordArray:
    """Data model of the 'time_weather' axis (time axis in the weather_xds
    subdataset when not interpolated to the main time axis. See also
    :py:class:`TimeCoordArray`."""

    data: Data[TimeWeather, float]
    """
    Time, expressed in seconds since the epoch (see ``scale`` &
    ``format``).
    """

    type: Attr[Time] = "time_weather"
    """ Coordinate type. Should be ``"time_weather"``. """

    units: Attr[UnitsSeconds] = ("s",)
    """ Units to associate with axis"""

    scale: Attr[AllowedTimeScales] = "utc"
    """ Astropy time scales, see :py:class:`TimeArray` """

    format: Attr[AllowedTimeFormats] = "unix"
    """ Astropy format, see :py:class:`TimeArray`"""


# For now allowing both some of the casacore frames (from "REST" to "TOPO" - all in uppercase) as well as
# the astropy frames (all in lowercase, taken from the list of SpectralCoord:
# https://docs.astropy.org/en/stable/coordinates/spectralcoord.html)
AllowedSpectralCoordFrames = Literal[
    "REST",
    # "LSRK" -> "lsrk",
    # "LSRD" -> "lsrd",
    "BARY",
    "GEO",
    "TOPO",
    # astropy frames
    "gcrs",
    "icrs",
    "hcrs",
    "lsrk",
    "lsrd",
    "lsr",
]


@xarray_dataarray_schema
class SpectralCoordArray:
    """
    Measures array for data variables and attributes that are spectral coordinates.
    """

    data: Data[ZD, float]

    units: Attr[UnitsHertz] = ("Hz",)

    observer: Attr[AllowedSpectralCoordFrames] = "gcrs"
    """
    Capitalized reference observers are from casacore. TOPO implies creating astropy earth_location.
    Astropy velocity reference frames are lowercase. Note that Astropy does not use the name 'TOPO' (telescope centric)
    rather it assumes if no velocity frame is given that this is the default.
    """

    type: Attr[SpectralCoord] = "spectral_coord"


AllowedLocationFrames = Literal["ITRF", "GRS80", "WGS84", "WGS72", "Undefined"]


AllowedLocationCoordinateSystems = Literal[
    "geocentric",
    "planetcentric",
    "geodetic",
    "planetodetic",
    "orbital",
]


@xarray_dataarray_schema
class LocationArray:
    """
    Measure type used for example in antenna_xds/ANTENNA_POSITION, field_and_source_xds/OBSERVER_POSITION
    Data dimensions can be EllipsoidPosLabel or CartesianPosLabel
    """

    data: Data[Union[EllipsoidPosLabel, CartesianPosLabel], float]

    units: Attr[UnitsOfLocationInMetersOrRadians]
    """
    If the units are a list of strings then it must be the same length as
    the last dimension of the data array. This allows for having different
    units in the same data array,for example geodetic coordinates could use
    ``['rad','rad','m']``.
    """

    frame: Attr[AllowedLocationFrames]
    """
    Can be ITRF, GRS80, WGS84, WGS72, Undefined
    """

    coordinate_system: Attr[AllowedLocationCoordinateSystems]
    """ Can be ``geocentric/planetcentric, geodetic/planetodetic, orbital`` """

    origin_object_name: Attr[str]
    """
    earth/sun/moon/etc
    """

    type: Attr[Location] = "location"
    """ """


@xarray_dataarray_schema
class EllipsoidPosLocationArray:
    """
    Measure type used for example in field_and_source_xds/SUB_OBSERVER_POSITION, SUB_SOLAR_POSITION
    """

    data: Data[EllipsoidPosLabel, float]

    frame: Attr[AllowedLocationFrames]
    """
    Can be ITRF, GRS80, WGS84, WGS72
    """

    coordinate_system: Attr[AllowedLocationCoordinateSystems]
    """ Can be ``geocentric/planetcentric, geodetic/planetodetic, orbital`` """

    origin_object_name: Attr[str]
    """
    earth/sun/moon/etc
    """

    type: Attr[Location] = "location"
    """ """

    units: Attr[UnitsOfPositionInRadians] = ("rad", "rad", "m")
    """
    If the units are a list of strings then it must be the same length as
    the last dimension of the data array. This allows for having different
    units in the same data array,for example geodetic coordinates could use
    ``['rad','rad','m']``.
    """


@xarray_dataarray_schema
class BaselineArray:
    """Model of the baseline_id coordinate in the main dataset (interferometric data, :py:class:`VisibiiltyXds`)"""

    data: Data[BaselineId, Union[numpy.int64, numpy.int32]]
    """Unique id for each baseline."""
    long_name: Optional[Attr[str]] = "Baseline ID"


@xarray_dataarray_schema
class BaselineAntennaNameArray:
    """Array of antenna_name by baseline_id, as used in main_xds and main_sd_xds
    (antenna_name by baseline_id dim"""

    data: Data[BaselineId, str]
    """Unique id for each baseline."""
    long_name: Optional[Attr[str]] = "Antenna name by baseline_id"


@xarray_dataarray_schema
class AntennaNameArray:
    """
    Model of the antenna_name coordinate, used in the main dataset (single dish data, :py:class:`VisibiiltyXds`)
    and several sub-datasets such as antenna_xds, pointing_xds, weather_xds, system_calibration_xds, gain_curve_xds, etc.
    """

    data: Data[AntennaName, str]
    """Unique name for each antenna(_station)."""
    long_name: Optional[Attr[str]] = "Antenna name"


AllowedDopplerTypes = Literal[
    "radio", "optical", "z", "ratio", "true", "relativistic", "beta", "gamma"
]


@xarray_dataarray_schema
class DopplerArray:
    """Doppler measure information for the frequency coordinate"""

    data: Data[ZD, numpy.float64]

    type: Attr[Doppler] = "doppler"
    """ Coordinate type. Should be ``"spectral_coord"``. """

    units: Attr[UnitsOfDopplerShift] = ("m/s",)
    """ Units to associate with axis, [ratio]/[m/s]"""

    doppler_type: Attr[AllowedDopplerTypes] = "radio"
    """
    Allowable values: radio, optical, z, ratio, true, relativistic, beta, gamma.
    Astropy only has radio and optical. Using casacore types: https://casadocs.readthedocs.io/en/stable/notebooks/memo-series.html?highlight=Spectral%20Frames#Spectral-Frames
    """


@xarray_dataarray_schema
class FrequencyArray:
    """Frequency coordinate in the main dataset."""

    data: Data[Frequency, float]
    """ Time, expressed in SI seconds since the epoch. """
    spectral_window_name: Attr[str]
    """ Name associated with spectral window. """
    frequency_group_name: Optional[Attr[str]]
    """ Name associated with frequency group - needed for multi-band VLBI fringe-fitting."""
    reference_frequency: Attr[SpectralCoordArray]
    """ A frequency representative of the spectral window, usually the sky
    frequency corresponding to the DC edge of the baseband. Used by the calibration
    system if a ﬁxed scaling frequency is required or in algorithms to identify the
    observing band. """
    channel_width: Attr[
        QuantityInHertzArray
    ]  # Not SpectralCoord, as it is a difference
    """ The nominal channel bandwidth. Same units as data array (see units key). """
    doppler: Optional[Attr[DopplerArray]]
    """ Doppler tracking information """

    type: Attr[SpectralCoord] = "spectral_coord"
    """ Coordinate type. Should be ``"spectral_coord"``. """
    long_name: Optional[Attr[str]] = "Frequency"
    """ Long-form name to use for axis"""
    units: Attr[UnitsHertz] = ("Hz",)
    """ Units to associate with axis"""
    observer: Attr[AllowedSpectralCoordFrames] = "icrs"
    """
    Astropy velocity reference frames (see :external:ref:`astropy-spectralcoord`).
    Note that Astropy does not use the name
    'topo' (telescope centric) velocity frame, rather it assumes if no velocity
    frame is given that this is the default.
    """


@xarray_dataarray_schema
class FrequencyCalArray:
    """The frequency_cal coordinate of the system calibration dataset. It has
    only measures data, as opposed to the frequency array of the main dataset."""

    data: Data[FrequencyCal, float]
    """ Time, expressed in SI seconds since the epoch. """

    type: Attr[SpectralCoord] = "spectral_coord"
    units: Attr[UnitsHertz] = ("Hz",)
    """ Units to associate with axis"""

    observer: Attr[AllowedSpectralCoordFrames] = "icrs"
    """
    Astropy velocity reference frames (see :external:ref:`astropy-spectralcoord`).
    Note that Astropy does not use the name
    'topo' (telescope centric) velocity frame, rather it assumes if no velocity
    frame is given that this is the default.
    """


@xarray_dataarray_schema
class PolarizationArray:
    """
    Possible correlations that can be formed from polarised receptors. Possible
    values, taken from `Measures/Stokes.h
    <https://github.com/casacore/casacore/blob/5a8df94738bdc36be27e695d7b14fe949a1cc2df/measures/Measures/Stokes.h>`_:

    * ``I``, ``Q``, ``U``, ``V`` (standard stokes parameters)
    * ``RR``, ``RL``, ``LR``, ``LL`` (circular correlation products)
    * ``XX``, ``XY``, ``YX``, ``YY`` (linear correlation products)
    * ``RX``, ``RY``, ``LX``, ``LY``, ``XR``, ``XL``, ``YR``, ``YL`` (mixed correlation products)
    * ``PP``, ``PQ``, ``QP``, ``QQ`` (general quasi-orthogonal correlation products)
    * ``RCircular``, ``LCircular``, ``Linear`` (single dish polarization types)
    * ``Ptotal`` (polarized intensity: ``sqrt(Q²+U²+V²)``)
    * ``Plinear`` (linearly polarized intensity: ``sqrt(Q²+U²)``)
    * ``PFtotal`` (polarization fraction: ``Ptotal/I``)
    * ``PFlinear`` (linear polarization fraction: ``Plinear/I``)
    * ``Pangle`` (linear polarization angle: ``0.5 arctan(U/Q)`` in radians)

    """

    data: Data[Polarization, str]
    """ Polarization names. """
    long_name: Optional[Attr[str]] = "Polarization"
    """ Long-form name to use for axis. Should be ``"Polarization"``"""


@xarray_dataarray_schema
class UvwLabelArray:
    """
    Coordinate axis to make up ``("u", "v", "w")`` tuple, see :py:class:`UvwArray`.
    """

    data: Data[UvwLabel, str] = ("u", "v", "w")
    """Should be ``('u','v','w')``, used by :py:class:`UvwArray`"""
    long_name: Optional[Attr[str]] = "U/V/W label"
    """ Long-form name to use for axis. Should be ``"U/V/W label"``"""


# Data variables
@xarray_dataarray_schema
class FlagArray:
    """
    An array of Boolean values with the same shape as `VISIBILITY`,
    representing the cumulative flags applying to this data matrix. Data are
    flagged bad if the ``FLAG`` array element is ``True``.
    """

    data: Data[
        Union[
            tuple[Time, BaselineId, Frequency, Polarization],
            tuple[Time, BaselineId, Frequency],
            tuple[Time, BaselineId],
            tuple[Time, AntennaName, Frequency, Polarization],  # SD
        ],
        bool,
    ]
    time: Coordof[TimeCoordArray]
    baseline_id: Optional[Coordof[BaselineArray]]  # Only IF
    antenna_name: Optional[Coordof[AntennaNameArray]]  # Only SD
    frequency: Coordof[FrequencyArray]
    polarization: Optional[Coordof[PolarizationArray]] = None
    long_name: Optional[Attr[str]] = "Visibility flags"

    allow_mutiple_versions: Optional[Attr[bool]] = True


@xarray_dataarray_schema
class WeightArray:
    """
    The weight for each channel, with the same shape as the associated
    :py:class:`VisibilityArray`, as assigned by the correlator or processor.

    Weight spectrum in ms v2 is renamed weight. Should be calculated as
    1/sigma^2 (sigma rms noise).
    """

    data: Data[
        Union[
            tuple[Time, BaselineId, Frequency, Polarization],
            tuple[Time, BaselineId, Frequency],
            tuple[Time, BaselineId],
            tuple[Time, AntennaName, Frequency, Polarization],  # SD
        ],
        Union[numpy.float16, numpy.float32, numpy.float64],
    ]
    """Visibility weights"""
    time: Coordof[TimeCoordArray]
    baseline_id: Optional[Coordof[BaselineArray]]  # Only IF
    antenna_name: Optional[Coordof[AntennaNameArray]]  # Only SD
    frequency: Coordof[FrequencyArray] = None
    polarization: Optional[Coordof[PolarizationArray]] = None
    long_name: Optional[Attr[str]] = "Visibility weights"

    allow_mutiple_versions: Optional[Attr[bool]] = True


# J2000=>fk5 is used most often. icrs is used less often. Both fk5 and icrs are also borrowed from the field center (to fix
# ITRF=>J2000). APP has only been seen in WSRT datasets.
AllowedUvwFrames = Literal[
    "fk5",
    "icrs",
    "APP",  # "apparent geocentric position", used in WSRT datasets
]


@xarray_dataarray_schema
class UvwArray:
    """
    Coordinates for the baseline from ``baseline_antenna2_id`` to
    ``baseline_antenna1_id``, i.e. the baseline is equal to the difference
    ``POSITION2 - POSITION1``. The UVW given are for the ``TIME_CENTROID``, and
    correspond in general to the reference type for the
    ``field_info.phase_dir``.

    The baseline direction should be: ``W`` towards source direction; ``V`` in
    plane through source and system's pole; ``U`` in direction of increasing
    longitude coordinate.  So citing
    http://casa.nrao.edu/Memos/CoordConvention.pdf: Consider an XYZ Celestial
    coordinate system centered at the location of the interferometer, with
    :math:`X` towards the East, :math:`Z` towards the NCP and :math:`Y` to
    complete a right-handed system. The UVW coordinate system is then defined
    by the hour-angle and declination of the phase-reference direction such
    that

    #. when the direction of observation is the NCP (`ha=0,dec=90`),
       the UVW coordinates are aligned with XYZ,
    #. V, W and the NCP are always on a Great circle,
    #. when W is on the local meridian, U points East
    #. when the direction of observation is at zero declination, an
       hour-angle of -6 hours makes W point due East.

    This definition also determines the sign of the phase of ``VISIBILITY``.

    """

    data: Data[
        Union[tuple[Time, BaselineId, UvwLabel]],
        Union[
            numpy.float16,
            numpy.float32,
            numpy.float64,
        ],
    ]
    """Baseline coordinates from ``baseline_antenna2_id`` to ``baseline_antenna1_id``"""
    time: Coordof[TimeCoordArray]
    baseline_id: Optional[Coordof[BaselineArray]]
    uvw_label: Coordof[UvwLabelArray] = ("u", "v", "w")

    long_name: Optional[Attr[str]] = "Baseline coordinates"
    """ Long-form name to use for axis. Should be ``"Baseline coordinates``"""

    type: Attr[Literal["uvw"]] = "uvw"
    frame: Attr[AllowedUvwFrames] = "icrs"
    """ To be defined in astropy (see for example https://github.com/astropy/astropy/issues/7766) """
    units: Attr[UnitsMeters] = ("m",)

    allow_mutiple_versions: Optional[Attr[bool]] = True


@xarray_dataarray_schema
class TimeSamplingArray:
    """
    Model of arrays of measures used in the main dataset for data variables such as TIME_CENTROID and
    TIME_CENTROID_EXTRA_PRECISION.
    """

    data: Data[
        Union[
            tuple[Time, BaselineId],
            tuple[Time, AntennaName],  # SD
        ],
        float,
    ]

    time: Coordof[TimeCoordArray]
    baseline_id: Optional[Coordof[BaselineArray]]  # Only IF
    antenna_name: Optional[Coordof[AntennaNameArray]]  # Only SD

    scale: Attr[AllowedTimeScales] = "utc"
    """ Astropy time scales, see :py:class:`astropy.time.Time` """
    format: Attr[AllowedTimeFormats] = "unix"
    """ Astropy format, see :py:class:`astropy.time.Time`. Default seconds from 1970-01-01 00:00:00 UTC """

    long_name: Optional[Attr[str]] = "Time sampling data"
    units: Attr[UnitsSeconds] = ("s",)


@xarray_dataarray_schema
class FreqSamplingArray:
    """
    Model of frequency related data variables of the main dataset, such as EFFECTIV_CHANNEL_WIDTH and FREQUENCY_CENTROID.
    """

    data: Data[
        Union[
            tuple[Time, BaselineId, Frequency, Polarization],
            tuple[Time, BaselineId, Frequency],
            tuple[Time, Frequency],
            tuple[Frequency],
        ],
        float,
    ]
    """
    Data about frequency sampling, such as centroid or integration
    time. Concrete function depends on concrete data array within
    :py:class:`VisibilityXds` or :py:class:`SpectrumXds`.
    """
    frequency: Coordof[FrequencyArray]
    time: Optional[Coordof[TimeCoordArray]] = None
    baseline_id: Optional[Coordof[BaselineArray]] = None
    polarization: Optional[Coordof[PolarizationArray]] = None
    long_name: Optional[Attr[str]] = "Frequency sampling data"
    units: Attr[UnitsHertz] = ("Hz",)
    observer: Attr[AllowedSpectralCoordFrames] = "icrs"
    """
    Astropy velocity reference frames (see :external:ref:`astropy-spectralcoord`).
    Note that Astropy does not use the name
    'topo' (telescope centric) velocity frame, rather it assumes if no velocity
    frame is given that this is the default.
    """


# Define FieldAndSourceXds dataset already here, as it is needed in the
# definition of VisibilityArray
@xarray_dataset_schema
class FieldSourceXds:
    """
    Field positions for each source.

    Defines a field position on the sky. For interferometers, this is the correlated field position.
    For single dishes, this is the nominal pointing direction.
    """

    source_name: Optional[Coord[Union[ZD, Time], str]]
    """ Source name. """
    field_name: Optional[Coord[Union[ZD, Time], str]]
    """Field name."""

    time: Optional[Coordof[TimeInterpolatedCoordArray]]
    """Midpoint of time for which this set of parameters is accurate. Labeled 'time' when interpolated to main time """
    time_ephemeris: Optional[Coordof[TimeEphemerisCoordArray]]
    """Midpoint of time for which this set of parameters is accurate. Labeled 'time_ephemeris' when not interpolating to main time """

    line_label: Optional[Coord[LineLabel, str]]
    """ Line labels (for line names and variables). """

    line_names: Optional[
        Coord[
            Union[
                tuple[LineLabel],
                tuple[Time, LineLabel],
                tuple[TimeEphemeris, LineLabel],
            ],
            str,
        ]
    ]
    """ Line names (e.g. v=1, J=1-0, SiO). """

    FIELD_PHASE_CENTER: Optional[
        Data[Union[ZD, tuple[Time], tuple[TimeEphemeris]], SkyCoordArray]
    ]
    """
    Offset from the SOURCE_DIRECTION that gives the direction of phase
    center for which the fringes have been stopped-that is a point source in
    this direction will produce a constant measured phase (page 2 of
    https://articles.adsabs.harvard.edu/pdf/1999ASPC..180...79F). For
    conversion from MSv2, frame refers column keywords by default. If frame
    varies with field, it refers DelayDir_Ref column instead.
    """

    FIELD_REFERENCE_CENTER: Optional[
        Data[Union[ZD, tuple[Time], tuple[TimeEphemeris]], SkyCoordArray]
    ]
    """
    Used in single-dish to record the associated reference direction if positionswitching
    been applied. For conversion from MSv2, frame refers column keywords by default. If
    frame varies with field, it refers DelayDir_Ref column instead.
    """

    SOURCE_LOCATION: Optional[
        Data[
            Union[
                ZD,
                tuple[Time],
                tuple[TimeEphemeris],
            ],
            SkyCoordArray,
        ]
    ]
    """
    CASA Table Cols: RA,DEC,Rho."Astrometric RA and Dec and Geocentric
    distance with respect to the observer’s location (Geocentric). "Adjusted
    for light-time aberration only. With respect to the reference plane and
    equinox of the chosen system (ICRF or FK4/B1950). If the FK4/B1950 frame
    output is selected, elliptic aberration terms are added. Astrometric RA/DEC
    is generally used when comparing or reducing data against a star catalog."
    https://ssd.jpl.nasa.gov/horizons/manual.html : 1. Astrometric RA & DEC
    """

    LINE_REST_FREQUENCY: Optional[
        Data[
            Union[
                tuple[LineLabel],
                tuple[Time, LineLabel],
                tuple[TimeEphemeris, LineLabel],
            ],
            SpectralCoordArray,
        ]
    ]
    """ Rest frequencies for the transitions. """

    LINE_SYSTEMIC_VELOCITY: Optional[
        Data[
            Union[
                tuple[LineLabel],
                tuple[Time, LineLabel],
                tuple[TimeEphemeris, LineLabel],
            ],
            QuantityInMetersPerSecondArray,
        ]
    ]
    """ Systemic velocity at reference """

    SOURCE_RADIAL_VELOCITY: Optional[
        Data[
            Union[ZD, tuple[Time], tuple[TimeEphemeris]], QuantityInMetersPerSecondArray
        ]
    ]
    """ CASA Table Cols: RadVel. Geocentric distance rate """

    NORTH_POLE_POSITION_ANGLE: Optional[
        Data[Union[ZD, tuple[Time], tuple[TimeEphemeris]], QuantityInRadiansArray]
    ]
    """ CASA Table cols: NP_ang, "Targets' apparent north-pole position angle (counter-clockwise with respect to direction of true-of-date reference-frame north pole) and angular distance from the sub-observer point (center of disc) at print time. A negative distance indicates the north-pole is on the hidden hemisphere." https://ssd.jpl.nasa.gov/horizons/manual.html : 17. North pole position angle & distance from disc center. """

    NORTH_POLE_ANGULAR_DISTANCE: Optional[
        Data[Union[ZD, tuple[Time], tuple[TimeEphemeris]], QuantityInRadiansArray]
    ]
    """ CASA Table cols: NP_dist, "Targets' apparent north-pole position angle (counter-clockwise with respect to direction of true-of date reference-frame north pole) and angular distance from the sub-observer point (center of disc) at print time. A negative distance indicates the north-pole is on the hidden hemisphere."https://ssd.jpl.nasa.gov/horizons/manual.html : 17. North pole position angle & distance from disc center. """

    SUB_OBSERVER_DIRECTION: Optional[
        Data[
            Union[
                ZD,
                tuple[Time],
                tuple[TimeEphemeris],
            ],
            EllipsoidPosLocationArray,
        ]
    ]
    """ CASA Table cols: DiskLong, DiskLat. "Apparent planetodetic longitude and latitude of the center of the target disc seen by the OBSERVER at print-time. This is not exactly the same as the "nearest point" for a non-spherical target shape (since the center of the disc might not be the point closest to the observer), but is generally very close if not a very irregular body shape. The IAU2009 rotation models are used except for Earth and MOON, which use higher-precision models. For the gas giants Jupiter, Saturn, Uranus and Neptune, IAU2009 longitude is based on the "System III" prime meridian rotation angle of the magnetic field. By contrast, pole direction (thus latitude) is relative to the body dynamical equator. There can be an offset between the magnetic pole and the dynamical pole of rotation. Down-leg light travel-time from target to observer is taken into account. Latitude is the angle between the equatorial plane and perpendicular to the reference ellipsoid of the body and body oblateness thereby included. The reference ellipsoid is an oblate spheroid with a single flatness coefficient in which the y-axis body radius is taken to be the same value as the x-axis radius. Whether longitude is positive to the east or west for the target will be indicated at the end of the output ephemeris." https://ssd.jpl.nasa.gov/horizons/manual.html : 14. Observer sub-longitude & sub-latitude """

    SUB_SOLAR_POSITION: Optional[
        Data[
            Union[
                ZD,
                tuple[Time],
                tuple[TimeEphemeris],
            ],
            EllipsoidPosLocationArray,
        ]
    ]
    """ CASA Table cols: Sl_lon, Sl_lat, r. "Heliocentric distance along with "Apparent sub-solar longitude and latitude of the Sun on the target. The apparent planetodetic longitude and latitude of the center of the target disc as seen from the Sun, as seen by the observer at print-time.  This is _NOT_ exactly the same as the "sub-solar" (nearest) point for a non-spherical target shape (since the center of the disc seen from the Sun might not be the closest point to the Sun), but is very close if not a highly irregular body shape.  Light travel-time from Sun to target and from target to observer is taken into account.  Latitude is the angle between the equatorial plane and the line perpendicular to the reference ellipsoid of the body. The reference ellipsoid is an oblate spheroid with a single flatness coefficient in which the y-axis body radius is taken to be the same value as the x-axis radius. Uses IAU2009 rotation models except for Earth and Moon, which uses a higher precision models. Values for Jupiter, Saturn, Uranus and Neptune are Set III, referring to rotation of their magnetic fields.  Whether longitude is positive to the east or west for the target will be indicated at the end of the output ephemeris." https://ssd.jpl.nasa.gov/horizons/manual.html : 15. Solar sub-longitude & sub-latitude  """

    HELIOCENTRIC_RADIAL_VELOCITY: Optional[
        Data[
            Union[ZD, tuple[Time], tuple[TimeEphemeris]], QuantityInMetersPerSecondArray
        ]
    ]
    """ CASA Table cols: rdot."The Sun's apparent range-rate relative to the target, as seen by the observer. A positive "rdot" means the target was moving away from the Sun, negative indicates movement toward the Sun." https://ssd.jpl.nasa.gov/horizons/manual.html : 19. Solar range & range-rate (relative to target) """

    OBSERVER_PHASE_ANGLE: Optional[
        Data[Union[ZD, tuple[Time], tuple[TimeEphemeris]], QuantityInRadiansArray]
    ]
    """ CASA Table cols: phang.""phi" is the true PHASE ANGLE at the observers' location at print time. "PAB-LON" and "PAB-LAT" are the FK4/B1950 or ICRF/J2000 ecliptic longitude and latitude of the phase angle bisector direction; the outward directed angle bisecting the arc created by the apparent vector from Sun to target center and the astrometric vector from observer to target center. For an otherwise uniform ellipsoid, the time when its long-axis is perpendicular to the PAB direction approximately corresponds to lightcurve maximum (or maximum brightness) of the body. PAB is discussed in Harris et al., Icarus 57, 251-258 (1984)." https://ssd.jpl.nasa.gov/horizons/manual.html : Phase angle and bisector """

    OBSERVER_POSITION: Optional[Data[ZD, LocationArray]]
    """ Observer location. """

    # --- Attributes ---
    doppler_shift_velocity: Optional[Attr[UnitsOfDopplerShift]]
    """ Velocity definition of the Doppler shift, e.g., RADIO or OPTICAL velocity in m/s """

    source_model_url: Optional[Attr[str]]
    """URL to access source model"""
    ephemeris_name: Optional[Attr[str]]
    """The name of the ephemeris. For example DE430.

    This can be used with Astropy solar_system_ephemeris.set('DE430'), see
    https://docs.astropy.org/en/stable/coordinates/solarsystem.html.
    """
    is_ephemeris: Attr[bool] = False

    type: Attr[Literal["field_and_source"]] = "field_and_source"
    """
    Type of dataset.
    """

    # --- Optional coordinates ---
    sky_dir_label: Optional[Coord[SkyDirLabel, str]] = ("ra", "dec")
    """ Coordinate labels of sky directions (typically shape 2 and 'ra', 'dec') """
    sky_pos_label: Optional[Coord[SkyPosLabel, str]] = ("ra", "dec", "dist")
    """ Coordinate lables of sky positions (typically shape 3 and 'ra', 'dec', 'dist') """
    ellipsoid_pos_label: Optional[Coord[EllipsoidPosLabel, str]] = (
        "lon",
        "lat",
        "height",
    )
    """ Coordinate labels of geodetic earth location data (typically shape 3 and 'lon', 'lat', 'height')"""
    cartesian_pos_label: Optional[Coord[CartesianPosLabel, str]] = ("x", "y", "z")
    """ Coordinate labels of geocentric earth location data (typically shape 3 and 'x', 'y', 'z')"""


@xarray_dataarray_schema
class SpectrumArray:
    """Definition of xr.DataArray for SPECTRUM data (single dish)"""

    data: Data[
        tuple[Time, AntennaName, Frequency, Polarization],
        Union[numpy.float64, numpy.float32, numpy.float16],
    ]

    time: Coordof[TimeCoordArray]
    antenna_name: Coordof[AntennaNameArray]
    frequency: Coordof[FrequencyArray]
    polarization: Coordof[PolarizationArray]

    field_and_source_xds: Attr[FieldSourceXds]
    long_name: Optional[Attr[str]] = "Spectrum values"
    """ Long-form name to use for axis. Should be ``"Spectrum values"``"""
    units: Attr[list[str]] = ("Jy",)


@xarray_dataarray_schema
class VisibilityArray:
    """Visibility data array in main dataset (interferometric data, :py:class:`VisibiiltyXds`)"""

    data: Data[
        tuple[Time, BaselineId, Frequency, Polarization],
        Union[numpy.complex64, numpy.complex128],
    ]

    time: Coordof[TimeCoordArray]
    baseline_id: Coordof[BaselineArray]
    polarization: Coordof[PolarizationArray]
    frequency: Coordof[FrequencyArray]

    field_and_source_xds: Attr[FieldSourceXds]
    long_name: Optional[Attr[str]] = "Visibility values"
    """ Long-form name to use for axis. Should be ``"Visibility values"``"""
    units: Attr[list[str]] = ("Jy",)

    allow_mutiple_versions: Optional[Attr[bool]] = True


# Info dicts


@dict_schema
class PartitionInfoDict:
    # spectral_window_id: missing / remove for good?
    spectral_window_name: str
    """ Spectral window Name """
    # field_id: missing / probably remove for good?
    field_name: list[str]
    """ List of all field names """
    polarization_setup: list[str]
    """ List of polrization bases. """
    scan_number: list[int]
    """ List of scan numbers. """
    source_name: list[str]
    """ List of source names. """
    # source_id: mising / remove for good?
    intents: list[str]
    """ Infromation in obs_mode column of MSv2 State table. """
    taql: Optional[str]
    """ The taql query used if converted from MSv2. """
    line_name: list[str]
    """ Spectral line names """


@dict_schema
class ObservationInfoDict:
    observer: list
    """List of observer names."""
    project: str
    """Project Code/Project_UID"""
    release_date: str
    """Project release date. This is the date on which the data may become
    public. Format: YYYY-MM-DDTHH:mm:ss.SSS (ISO 8601)"""
    execution_block_id: Optional[str]
    """ ASDM: Indicates the position of the execution block in the project
    (sequential numbering starting at 1).  """
    execution_block_number: Optional[int]
    """ASDM: Indicates the position of the execution block in the project
    (sequential numbering starting at 1)."""
    execution_block_UID: Optional[str]
    """ASDM: The archive’s UID of the execution block."""
    session_reference: Optional[str]
    """ASDM: The observing session reference."""
    observing_script: Optional[str]
    """ASDM: The text of the observation script."""
    observing_script_UID: Optional[str]
    """ASDM: A reference to the Entity which contains the observing script."""
    observing_log: Optional[str]
    """ASDM: Logs of the observation during this execu- tion block."""


@dict_schema
class ProcessorInfoDict:
    type: str
    """Processor type; reserved keywords include (”CORRELATOR” -
    interferometric correlator; ”SPECTROMETER” - single-dish correlator;
    ”RADIOMETER” - generic detector/integrator)."""
    sub_type: str
    """Processor sub-type, e.g. ”GBT” or ”JIVE”."""


# Data Sets


@xarray_dataset_schema
class AntennaXds:
    """
    Antenna dataset: global antenna properties for each antenna.
    """

    # Coordinates
    antenna_name: Coordof[AntennaNameArray]
    """ Antenna name """
    station: Coord[AntennaName, str]
    """ Name of the station pad (relevant to arrays with moving antennas). """
    mount: Coord[AntennaName, str]
    """ Mount type of the antenna. Reserved keywords include: ”EQUATORIAL” - equatorial mount;
    ”ALT-AZ” - azimuth-elevation mount;
    "ALT-AZ+ROTATOR"  alt-az mount with feed rotator; introduced for ASKAP dishes;
    "ALT-AZ+NASMYTH-R": Nasmyth mount with receivers at the right-hand side of the cabin. Many high-frequency antennas used for VLBI have such a mount typel;
    "ALT-AZ+NASMYTH-L:: Nasmyth mount with receivers at the left-hand side of the cabin.
    ”X-Y” - x-y mount;
    ”SPACE-HALCA” - specific orientation model."""
    telescope_name: Coord[AntennaName, str]
    """ Useful when data is combined from mutiple arrays for example ACA + ALMA. """
    receptor_label: Coord[ReceptorLabel, str]
    """ Names of receptors """
    polarization_type: Coord[tuple[AntennaName, ReceptorLabel], str]
    """ Polarization type to which each receptor responds (e.g. ”R”,”L”,”X” or ”Y”).
    This is the receptor polarization type as recorded in the final correlated data (e.g. ”RR”); i.e.
    as measured after all polarization combiners. ['X','Y'], ['R','L'] """
    cartesian_pos_label: Optional[Coord[CartesianPosLabel, str]]
    """ (x,y,z) - either cartesian or ellipsoid """
    ellipsoid_pos_label: Optional[Coord[EllipsoidPosLabel, str]]
    """ (lon, lat, dist) - either cartesian or ellipsoid"""

    # Data variables
    ANTENNA_POSITION: Data[tuple[AntennaName], LocationArray]
    """
    In a right-handed frame, X towards the intersection of the equator and
    the Greenwich meridian, Z towards the pole.
    """
    ANTENNA_DISH_DIAMETER: Optional[Data[tuple[AntennaName], QuantityInMetersArray]]
    """
    The diameter of the main reflector (or the largest dimension for non-circular apertures).
    """
    ANTENNA_EFFECTIVE_DISH_DIAMETER: Optional[
        Data[tuple[AntennaName], QuantityInMetersArray]
    ]
    """ Effective dish diameter used in computing beam model (such as airy disk). """

    ANTENNA_BLOCKAGE: Optional[Data[tuple[AntennaName], QuantityInMetersArray]]
    """
    Blockage caused by secondary reflector used in computing beam model (such as airy disk).
    """

    # TODO: setting BEAM_OFFSET and RECEPTOR_ANGLE as optional for now, as it
    # is not present in some datasets (example: test_alma_ephemris_mosaic)
    ANTENNA_RECEPTOR_ANGLE: Optional[
        Data[tuple[AntennaName, ReceptorLabel], QuantityInRadiansArray]
    ]
    """
    Polarization reference angle. Converts into parallactic angle in the sky domain.
    """
    ANTENNA_FOCUS_LENGTH: Optional[Data[tuple[AntennaName], QuantityInMetersArray]]
    """
    Focus length. As defined along the optical axis of the antenna.
    """

    # Attributes
    overall_telescope_name: Optional[Attr[str]]
    """
    The name of the collection of arrays and dishes that were used for the observation.
    In many instances this will only be a single array or dish. An example of a
    telescope consistening of mutiple arrays and dishes is the EHT. The coordinate
    telescope_name will give the names of the constituent arrays and dishes. From
    MSv2 observation table.
    """
    relocatable_antennas: Optional[Attr[bool]]
    """ Can the antennas be moved (ALMA, VLA, NOEMA) """
    type: Attr[Literal["antenna"]] = "antenna"
    """
    Type of dataset. Expected to be ``antenna``
    """


@xarray_dataset_schema
class GainCurveXds:
    """
    Gain curve dataset. See See https://casacore.github.io/casacore-notes/265.pdf for a full description.
    """

    # Coordinates
    antenna_name: Coordof[AntennaNameArray]
    """ Antenna name """
    station: Coord[AntennaName, str]
    """ Name of the station pad (relevant to arrays with moving antennas). """
    mount: Coord[AntennaName, str]
    """ Mount type of the antenna. Reserved keywords include: ”EQUATORIAL” - equatorial mount;
    ”ALT-AZ” - azimuth-elevation mount;
    "ALT-AZ+ROTATOR"  alt-az mount with feed rotator; introduced for ASKAP dishes;
    "ALT-AZ+NASMYTH-R": Nasmyth mount with receivers at the right-hand side of the cabin. Many high-frequency antennas used for VLBI have such a mount typel;
    "ALT-AZ+NASMYTH-L:: Nasmyth mount with receivers at the left-hand side of the cabin.
    ”X-Y” - x-y mount;
    ”SPACE-HALCA” - specific orientation model."""
    telescope_name: Coord[AntennaName, str]
    """ Useful when data is combined from mutiple arrays for example ACA + ALMA. """
    receptor_label: Coord[ReceptorLabel, str]
    """ Names of receptors """
    polarization_type: Optional[Coord[tuple[AntennaName, ReceptorLabel], str]]
    """ Polarization type to which each receptor responds (e.g. ”R”,”L”,”X” or ”Y”).
    This is the receptor polarization type as recorded in the final correlated data (e.g. ”RR”); i.e.
    as measured after all polarization combiners. ['X','Y'], ['R','L'] """
    gain_curve_type: Optional[Coord[AntennaName, str]]
    """
    Gain curve type. Reserved keywords include:
    (”POWER(EL)” - Power as a function of elevation;
     ”POWER(ZA)” - Power as a function of zenith angle;
     ”VOLTAGE(EL)” - Voltage as a function of elevation;
     ”VOLTAGE(ZA)” - Voltage as a function of zenith angle). See https://casacore.github.io/casacore-notes/265.pdf
    """
    poly_term: Coord[PolyTerm, int]
    """Term orders in gain curve polynomial"""

    GAIN_CURVE: Data[tuple[AntennaName, PolyTerm, ReceptorLabel], numpy.float64]
    """ Coeﬃcients of the polynomial that describes the (power or voltage) gain.  """
    GAIN_CURVE_INTERVAL: Data[tuple[AntennaName], QuantityInSecondsArray]
    """ Time interval. """
    GAIN_CURVE_SENSITIVITY: Data[
        tuple[AntennaName, ReceptorLabel], QuantityInKelvinPerJanskyArray
    ]
    """ Sensitivity of the antenna expressed in K/Jy. This is what AIPS calls “DPFU”. """

    measured_date: Attr[str]
    """
    Date gain curve was measured. Format: YYYY-MM-DDTHH:mm:ss.SSS (ISO 8601)
    """
    type: Attr[Literal["gain_curve"]] = "gain_curve"
    """
    Type of dataset. Expected to be ``gain_curve``
    """


@xarray_dataset_schema
class PhaseCalibrationXds:
    """
    Phase calibration dataset: signal chain phase calibration measurements.
    """

    # Coordinates
    antenna_name: Coordof[AntennaNameArray]
    """ Antenna name """
    station: Coord[AntennaName, str]
    """ Name of the station pad (relevant to arrays with moving antennas). """
    mount: Coord[AntennaName, str]
    """ Mount type of the antenna. Reserved keywords include: ”EQUATORIAL” - equatorial mount;
    ”ALT-AZ” - azimuth-elevation mount;
    "ALT-AZ+ROTATOR"  alt-az mount with feed rotator; introduced for ASKAP dishes;
    "ALT-AZ+NASMYTH-R": Nasmyth mount with receivers at the right-hand side of the cabin. Many high-frequency antennas used for VLBI have such a mount typel;
    "ALT-AZ+NASMYTH-L:: Nasmyth mount with receivers at the left-hand side of the cabin.
    ”X-Y” - x-y mount;
    ”SPACE-HALCA” - specific orientation model."""
    telescope_name: Coord[AntennaName, str]
    """ Useful when data is combined from mutiple arrays for example ACA + ALMA. """
    receptor_label: Coord[ReceptorLabel, str]
    """ Names of receptors """
    polarization_type: Optional[Coord[tuple[AntennaName, ReceptorLabel], str]]
    """ Polarization type to which each receptor responds (e.g. ”R”,”L”,”X” or ”Y”).
    This is the receptor polarization type as recorded in the final correlated data (e.g. ”RR”); i.e.
    as measured after all polarization combiners. ['X','Y'], ['R','L'] """
    time: Optional[Coordof[TimeInterpolatedCoordArray]]
    """ Time for VLBI phase cal"""
    time_phase_cal: Optional[Coord[TimePhaseCal, numpy.float64]]
    """ Time for VLBI phase cal"""
    tone_label: Optional[Coord[ToneLabel, str]]
    """
    Phase-cal tones that are measured. This number may vary by antenna, and may vary by spectral window as well, especially
    if spectral windows of varying widths are supported
    """

    PHASE_CAL: Data[
        Union[
            tuple[AntennaName, Time, ReceptorLabel, ToneLabel],
            tuple[AntennaName, TimePhaseCal, ReceptorLabel, ToneLabel],
        ],
        numpy.complex64,
    ]
    """
    Phase calibration measurements. These are provided as complex values that represent both the phase
and amplitude for a measured phase-cal tone. Measurements are provided as a two-dimensional array such that
separate measurements can be provided for each receptor of a feed (so separate values for each polarization)
for each of the measured tones. See https://casacore.github.io/casacore-notes/265.pdf
    """
    PHASE_CAL_CABLE_CAL: Data[
        Union[tuple[AntennaName, Time], tuple[AntennaName, TimePhaseCal]],
        QuantityInSecondsArray,
    ]
    """
    Cable calibration measurement. This is a measurement of the delay in the cable that provides the
reference signal to the receiver. There should be only a single reference signal per feed (even if that feed has
multiple receptors) so this is provided as a simple scalar. See https://casacore.github.io/casacore-notes/265.pdf
    """
    PHASE_CAL_INTERVAL: Data[
        Union[tuple[AntennaName, Time], tuple[AntennaName, TimePhaseCal]],
        QuantityInSecondsArray,
    ]
    """
    Time interval. See https://casacore.github.io/casacore-notes/265.pdf
    """
    PHASE_CAL_TONE_FREQUENCY: Data[
        Union[
            tuple[AntennaName, Time, ReceptorLabel, ToneLabel],
            tuple[AntennaName, TimePhaseCal, ReceptorLabel, ToneLabel],
        ],
        QuantityInHertzArray,
    ]
    """
    The sky frequencies of each measured phase-cal tone. See https://casacore.github.io/casacore-notes/265.pdf
    """

    type: Attr[Literal["phase_calibration"]] = "phase_calibration"
    """
    Type of dataset. Expected to be ``phase_calibration``
    """


@xarray_dataset_schema
class WeatherXds:
    """
    Weather dataset: station positions and time-dependent mean external atmosphere and weather information
    """

    # Coordinates
    station_name: Coord[StationName, str]
    """ Station identifier """
    time: Optional[Coordof[TimeInterpolatedCoordArray]]
    """ Mid-point of the time interval. Labeled 'time' when interpolated to main time axis """
    time_weather: Optional[Coordof[TimeWeatherCoordArray]]
    """ Mid-point of the time interval. Labeled 'time_cal' when not interpolated to main time axis """
    antenna_name: Optional[Coordof[AntennaNameArray]]
    """ Antenna identifier """
    ellipsoid_pos_label: Optional[Coord[EllipsoidPosLabel, str]] = (
        "lon",
        "lat",
        "height",
    )
    """ Coordinate labels of geodetic earth location data (typically shape 3 and 'lon', 'lat', 'height')"""
    cartesian_pos_label: Optional[Coord[CartesianPosLabel, str]] = ("x", "y", "z")
    """ Coordinate labels of geocentric earth location data (typically shape 3 and 'x', 'y', 'z')"""

    # Data variables (all optional)
    H2O: Optional[
        Data[
            Union[tuple[StationName, Time], tuple[StationName, TimeWeather]],
            QuantityInPerSquareMetersArray,
        ]
    ] = None
    """ Average column density of water """
    IONOS_ELECTRON: Optional[
        Data[
            Union[
                tuple[StationName, Time],
                tuple[StationName, TimeWeather],
            ],
            QuantityInPerSquareMetersArray,
        ]
    ] = None
    """ Average column density of electrons """
    PRESSURE: Optional[
        Data[
            Union[
                tuple[StationName, Time],
                tuple[StationName, TimeWeather],
            ],
            QuantityInPascalArray,
        ]
    ] = None
    """ Ambient atmospheric pressure """
    REL_HUMIDITY: Optional[
        Data[
            Union[
                tuple[StationName, Time],
                tuple[StationName, TimeWeather],
            ],
            numpy.float64,
        ]
    ] = None
    """ Ambient relative humidity """
    TEMPERATURE: Optional[
        Data[
            Union[
                tuple[StationName, Time],
                tuple[StationName, TimeWeather],
            ],
            QuantityInKelvinArray,
        ]
    ] = None
    """ Ambient air temperature for an antenna """
    DEW_POINT: Optional[
        Data[
            Union[
                tuple[StationName, Time],
                tuple[StationName, TimeWeather],
            ],
            QuantityInKelvinArray,
        ]
    ] = None
    """ Dew point """
    WIND_DIRECTION: Optional[
        Data[
            Union[
                tuple[StationName, Time],
                tuple[StationName, TimeWeather],
            ],
            QuantityInRadiansArray,
        ]
    ] = None
    """ Average wind direction """
    WIND_SPEED: Optional[
        Data[
            Union[
                tuple[StationName, Time],
                tuple[StationName, TimeWeather],
            ],
            QuantityInMetersPerSecondArray,
        ]
    ] = None
    """ Average wind speed """
    STATION_POSITION: Optional[Data[tuple[StationName], LocationArray]] = None
    """ Station position """

    # Attributes
    type: Attr[Literal["weather"]] = "weather"
    """
    Type of dataset.
    """


@xarray_dataset_schema
class PointingXds:
    """
    Pointing dataset: antenna pointing information.

    In the past the relationship and definition of the pointing infromation has not been clear. Here we attempt to clarify it by explaining the relationship between the ASDM, MSv2 and MSv4 pointing information.

    The following abreviations are used:

    - M2: Measurement Set version 2
    - M4: Measurement Set version 4
    - A : ASDM

    The following definitions come from the ASDM's `SDM Tables Short Description <https://drive.google.com/file/d/16a3g0GQxgcO7N_ZabfdtexQ8r2jRbYIS/view>`_ page 97-99:

    - A_encoder: The values measured from the antenna. They may be however affected by metrology, if applied. Note that for ALMA this column will contain positions obtained using the AZ POSN RSP and EL POSN RSP monitor points of the ACU and not the GET AZ ENC and GET EL ENC monitor points (as these do not include the metrology corrections). It is agreed that the the vendor pointing model will never be applied. AZELNOWAntenna.position
    - A_pointingDirection: This is the commanded direction of the antenna. It is obtained by adding the target and offset columns, and then applying the pointing model referenced by PointingModelId. The pointing model can be the composition of the absolute pointing model and of a local pointing model. In that case their coefficients will both be in the PointingModel table.
    - A_target: This is the field center direction (as given in the Field Table), possibly affected by the optional antenna-based sourceOffset. This column is in horizontal coordinates. AZELNOWAntenna.position
    - A_offset: Additional offsets in horizontal coordinates (usually meant for measuring the pointing corrections, mapping the antenna beam, ...). AZELNOWAntenna.positiontarget
    - A_sourceOffset : Optionally, the antenna-based mapping offsets in the field. These are in the equatorial system, and used, for instance, in on-the-fly mapping when the antennas are driven independently across the field.

    M2_DIRECTION = rotate(A_target,A_offset)   #A_target is rotated to by A_offset

    if withPointingCorrection : M2_DIRECTION = rotate(A_target,A_offset) + (A_encoder - A_pointingDirection)

    M2_TARGET = A_target
    M2_POINTING_OFFSET = A_offset
    M2_ENCODER = A_encoder

    It should be noted that definition of M2_direction is not consistent, it depends if withPointingCorrection is set to True or False (see the `importasdm documenation <https://casadocs.readthedocs.io/en/v6.2.0/api/tt/casatasks.data.importasdm.html#with-pointing-correction>`_  and `code <https://open-bitbucket.nrao.edu/projects/CASA/repos/casa6/browse/casatools/src/tools/sdm/sdm_cmpt.cc#2257>`_ for details).

    M4_DIRECTION = M2_DIRECTION (withPointingCorrection=True)
    M4_ENCODER = M2_ENCODER

    """

    antenna_name: Coordof[AntennaNameArray]
    """
    Antenna name, as specified by baseline_antenna1/2_name in visibility dataset
    """

    local_sky_dir_label: Coord[LocalSkyDirLabel, str]
    """
    Direction labels.
    """

    POINTING_BEAM: Data[
        Union[
            tuple[Time, AntennaName],
            tuple[TimePointing, AntennaName],
            tuple[Time, AntennaName, nPolynomial],
            tuple[TimePointing, AntennaName, nPolynomial],
        ],
        LocalSkyCoordArray,
    ]
    """
    The direction of the peak response of the beam and is equavalent to the MSv2 DIRECTION (M2_direction) with_pointing_correction=True, optionally expressed as polynomial coefficients.
    """

    time: Optional[Coordof[TimeInterpolatedCoordArray]] = None
    """
    Mid-point of the time interval for which the information in this row is
    valid. Required to use the same time measure reference as in visibility dataset.
    Labeled 'time' when interpolating to main time axis.
    """
    time_pointing: Optional[Coordof[TimePointingCoordArray]] = None
    """ Midpoint of time for which this set of parameters is accurate. Labeled
    'time_pointing' when not interpolating to main time axis """

    POINTING_DISH_MEASURED: Optional[
        Data[
            Union[
                tuple[Time, AntennaName],
                tuple[TimePointing, AntennaName],
            ],
            LocalSkyCoordArray,
        ]
    ] = None
    """
    The current encoder values on the primary axes of the mount type for
    the antenna. ENCODER in MSv2 (M2_encoder).
    """
    POINTING_OVER_THE_TOP: Optional[
        Data[Union[tuple[Time, AntennaName], tuple[TimePointing, AntennaName]], bool]
    ] = None
    """
    True if the antenna was driven to this position ”over the top” (az-el mount).
    """

    # Attributes
    type: Attr[Literal["pointing"]] = "pointing"
    """
    Type of dataset.
    """


@xarray_dataset_schema
class SystemCalibrationXds:
    """
    System calibration dataset: time- and frequency- variable calibration measurements for each antenna,
    as indexed on receptor
    """

    # Coordinates
    antenna_name: Coordof[AntennaNameArray]
    """ Antenna identifier """
    receptor_label: Coord[ReceptorLabel, numpy.int64]
    """  """
    time: Optional[Coordof[TimeInterpolatedCoordArray]] = None
    """ Midpoint of time for which this set of parameters is accurate. Labeled 'time' when interpolating to main time axis """
    time_cal: Optional[Coordof[TimeCalCoordArray]] = None
    """ Midpoint of time for which this set of parameters is accurate. Labeled 'time_cal' when not interpolating to main time axis """
    frequency: Optional[Coordof[FrequencyCalArray]] = None
    """  """
    frequency_cal: Optional[Coord[FrequencyCal, int]] = None
    """TODO: What is this?"""

    # Data variables (all optional)
    PHASE_DIFFERENCE: Optional[
        Data[
            Union[tuple[AntennaName, TimeCal], tuple[AntennaName, Time]],
            QuantityInRadiansArray,
        ]
    ] = None
    """ Phase difference between receptor 0 and receptor 1 """
    TCAL: Optional[
        Data[
            Union[
                tuple[AntennaName, TimeCal, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, TimeCal, ReceptorLabel, Frequency],
                tuple[AntennaName, TimeCal, ReceptorLabel],
                tuple[AntennaName, Time, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, Time, ReceptorLabel, Frequency],
                tuple[AntennaName, Time, ReceptorLabel],
            ],
            QuantityInKelvinArray,
        ]
    ] = None
    """ Calibration temp """
    TRX: Optional[
        Data[
            Union[
                tuple[AntennaName, TimeCal, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, TimeCal, ReceptorLabel, Frequency],
                tuple[AntennaName, TimeCal, ReceptorLabel],
                tuple[AntennaName, Time, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, Time, ReceptorLabel, Frequency],
                tuple[AntennaName, Time, ReceptorLabel],
            ],
            QuantityInKelvinArray,
        ]
    ] = None
    """ Receiver temperature """
    TSKY: Optional[
        Data[
            Union[
                tuple[AntennaName, TimeCal, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, TimeCal, ReceptorLabel, Frequency],
                tuple[AntennaName, TimeCal, ReceptorLabel],
                tuple[AntennaName, Time, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, Time, ReceptorLabel, Frequency],
                tuple[AntennaName, Time, ReceptorLabel],
            ],
            QuantityInKelvinArray,
        ]
    ] = None
    """ Sky temperature """
    TSYS: Optional[
        Data[
            Union[
                tuple[AntennaName, TimeCal, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, TimeCal, ReceptorLabel, Frequency],
                tuple[AntennaName, TimeCal, ReceptorLabel],
                tuple[AntennaName, Time, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, Time, ReceptorLabel, Frequency],
                tuple[AntennaName, Time, ReceptorLabel],
            ],
            QuantityInKelvinArray,
        ]
    ] = None
    """ System temperature """
    TANT: Optional[
        Data[
            Union[
                tuple[AntennaName, TimeCal, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, TimeCal, ReceptorLabel, Frequency],
                tuple[AntennaName, TimeCal, ReceptorLabel],
                tuple[AntennaName, Time, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, Time, ReceptorLabel, Frequency],
                tuple[AntennaName, Time, ReceptorLabel],
            ],
            QuantityInKelvinArray,
        ]
    ] = None
    """ Antenna temperature """
    TANT_SYS: Optional[
        Data[
            Union[
                tuple[AntennaName, TimeCal, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, TimeCal, ReceptorLabel, Frequency],
                tuple[AntennaName, TimeCal, ReceptorLabel],
                tuple[AntennaName, Time, ReceptorLabel, FrequencyCal],
                tuple[AntennaName, Time, ReceptorLabel, Frequency],
                tuple[AntennaName, Time, ReceptorLabel],
            ],
            QuantityInKelvinArray,
        ]
    ] = None
    """ TANT/TSYS """

    # Attributes
    type: Attr[Literal["system_calibration"]] = "system_calibration"
    """
    Type of dataset.
    """


@xarray_dataset_schema
class PhasedArrayXds:
    """Not specified. Not implemented."""

    pass


@xarray_dataset_schema
class DopplerXds:
    """Not specified. Not implemented."""

    pass


@xarray_dataset_schema
class VisibilityXds:
    """Main dataset for interferometric data"""

    # --- Required Coordinates ---
    time: Coordof[TimeCoordArray]
    """
    The time coordinate is the mid-point of the nominal sampling interval, as
    speciﬁed in the ``ms_v4.time.attrs['integration_time']`` (ms v2 interval).
    """
    baseline_id: Coordof[BaselineArray]
    """ Baseline ID """
    frequency: Coordof[FrequencyArray]
    """Center frequencies for each channel."""
    polarization: Coordof[PolarizationArray]
    """
    Labels for polarization types, e.g. ``['XX','XY','YX','YY']``, ``['RR','RL','LR','LL']``.
    """

    # --- Required data variables ---

    VISIBILITY: Dataof[VisibilityArray]
    """Complex visibilities, either simulated or measured by interferometer."""

    baseline_antenna1_name: Coordof[BaselineAntennaNameArray]
    """Antenna name for 1st antenna in baseline. Maps to ``attrs['antenna_xds'].antenna_name``"""
    baseline_antenna2_name: Coordof[BaselineAntennaNameArray]
    """Antenna name for 2nd antenna in baseline. Maps to ``attrs['antenna_xds'].antenna_name``"""

    # --- Required Attributes ---
    partition_info: Attr[PartitionInfoDict]
    observation_info: Attr[ObservationInfoDict]
    processor_info: Attr[ProcessorInfoDict]
    antenna_xds: Attr[AntennaXds]

    schema_version: Attr[str]
    """Semantic version of xradio data format"""
    creation_date: Attr[str]
    """Date visibility dataset was created . Format: YYYY-MM-DDTHH:mm:ss.SSS (ISO 8601)"""

    type: Attr[Literal["visibility"]] = "visibility"
    """
    Dataset type
    """

    # --- Optional Coordinates ---
    polarization_mixed: Optional[Coord[tuple[BaselineId, Polarization], str]] = None
    """
    If the polarizations are not constant over baseline
    """
    uvw_label: Optional[Coordof[UvwLabelArray]] = None
    """ u,v,w """
    scan_number: Optional[Coord[Time, Union[numpy.int64, numpy.int32]]] = None
    """Arbitary scan number to identify data taken in the same logical scan."""

    # --- Optional data variables / arrays ---

    # VISIBILITY_CORRECTED: Optional[Dataof[VisibilityArray]] = None
    # VISIBILITY_MODEL: Optional[Dataof[VisibilityArray]] = None

    FLAG: Dataof[FlagArray] = None
    WEIGHT: Dataof[WeightArray] = None
    UVW: Dataof[UvwArray] = None
    EFFECTIVE_INTEGRATION_TIME: Optional[
        Data[
            Union[
                tuple[Time, BaselineId],
                tuple[Time, BaselineId, Frequency],
                tuple[Time, BaselineId, Frequency, Polarization],
            ],
            QuantityInSecondsArray,
        ]
    ] = None
    """
    The integration time, including the effects of missing data, in contrast to
    ``integration_time`` attribute of the ``time`` coordinate,
    see :py:class:`TimeArray`. (MS v2: ``exposure``).
    """
    TIME_CENTROID: Optional[Dataof[TimeSamplingArray]] = None
    """
    The time centroid of the visibility, includes the effects of missing data
    unlike the ``time`` coordinate, see :py:class:`TimeArray`.
    """
    TIME_CENTROID_EXTRA_PRECISION: Optional[Dataof[TimeSamplingArray]] = None
    """Additional precision for ``TIME_CENTROID``"""
    EFFECTIVE_CHANNEL_WIDTH: Optional[Dataof[FreqSamplingArray]] = None
    """The channel bandwidth that includes the effects of missing data."""
    FREQUENCY_CENTROID: Optional[Dataof[FreqSamplingArray]] = None
    """Includes the effects of missing data unlike ``frequency``."""

    # --- Optional Attributes ---
    pointing_xds: Optional[Attr[PointingXds]] = None
    system_calibration_xds: Optional[Attr[SystemCalibrationXds]] = None
    gain_curve_xds: Optional[Attr[GainCurveXds]] = None
    phase_calibration_xds: Optional[Attr[PhaseCalibrationXds]] = None
    weather_xds: Optional[Attr[WeatherXds]] = None
    phased_array_xds: Optional[Attr[PhasedArrayXds]] = None

    xradio_version: Optional[Attr[str]] = None
    """ Version of XRADIO used if converted from MSv2. """

    intent: Optional[Attr[str]] = None
    """Identifies the intention of the scan, such as to calibrate or observe a
    target. See :ref:`scan intents` for possible values.
    """
    data_description_id: Optional[Attr[str]] = None
    """
    The id assigned to this combination of spectral window and polarization setup.
    """


@xarray_dataset_schema
class SpectrumXds:
    """Main dataset for single dish data"""

    # --- Required Coordinates ---
    time: Coordof[TimeCoordArray]
    """
    The time coordinate is the mid-point of the nominal sampling interval, as
    speciﬁed in the ``ms_v4.time.attrs['integration_time']`` (ms v2 interval).
    """
    antenna_name: Coordof[AntennaNameArray]
    """ antenna_name """
    frequency: Coordof[FrequencyArray]
    """Center frequencies for each channel."""
    polarization: Coordof[PolarizationArray]
    """
    Labels for polarization types, e.g. ``['XX','XY','YX','YY']``, ``['RR','RL','LR','LL']``.
    """

    # --- Required data variables ---
    SPECTRUM: Dataof[SpectrumArray]
    """Single dish data, either simulated or measured by an antenna."""

    # --- Required Attributes ---
    partition_info: Attr[PartitionInfoDict]
    observation_info: Attr[ObservationInfoDict]
    processor_info: Attr[ProcessorInfoDict]
    antenna_xds: Attr[AntennaXds]

    schema_version: Attr[str]
    """Semantic version of xradio data format"""
    creation_date: Attr[str]
    """Date MSv4 was created . Format: YYYY-MM-DDTHH:mm:ss.SSS (ISO 8601)"""

    type: Attr[Literal["spectrum"]] = "spectrum"
    """
    Dataset type
    """

    # --- Optional Coordinates ---
    polarization_mixed: Optional[Coord[tuple[AntennaName, Polarization], str]] = None
    """
    If the polarizations are not constant over baseline
    """
    scan_number: Optional[Coord[Time, Union[numpy.int64, numpy.int32]]] = None
    """Arbitary scan number to identify data taken in the same logical scan."""

    # SPECTRUM_CORRECTED: Optional[Dataof[SpectrumArray]] = None

    FLAG: Dataof[FlagArray] = None
    WEIGHT: Dataof[WeightArray] = None

    # --- Optional data variables / arrays ---
    EFFECTIVE_INTEGRATION_TIME: Optional[
        Data[
            Union[
                tuple[Time, AntennaName],
                tuple[Time, AntennaName, Frequency],
                tuple[Time, AntennaName, Frequency, Polarization],
            ],
            QuantityInSecondsArray,
        ]
    ] = None
    """
    The integration time, including the effects of missing data, in contrast to
    ``integration_time`` attribute of the ``time`` coordinate,
    see :py:class:`TimeArray`. (MS v2: ``exposure``).
    """
    TIME_CENTROID: Optional[Dataof[TimeSamplingArray]] = None
    """
    The time centroid of the visibility, includes the effects of missing data
    unlike the ``time`` coordinate, see :py:class:`TimeArray`.
    """
    TIME_CENTROID_EXTRA_PRECISION: Optional[Dataof[TimeSamplingArray]] = None
    """Additional precision for ``TIME_CENTROID``"""
    EFFECTIVE_CHANNEL_WIDTH: Optional[Dataof[FreqSamplingArray]] = None
    """The channel bandwidth that includes the effects of missing data."""
    FREQUENCY_CENTROID: Optional[Dataof[FreqSamplingArray]] = None
    """Includes the effects of missing data unlike ``frequency``."""

    # --- Optional Attributes ---
    pointing_xds: Optional[Attr[PointingXds]] = None
    system_calibration_xds: Optional[Attr[SystemCalibrationXds]] = None
    gain_curve_xds: Optional[Attr[GainCurveXds]] = None
    phase_calibration_xds: Optional[Attr[PhaseCalibrationXds]] = None
    weather_xds: Optional[Attr[WeatherXds]] = None
    phased_array_xds: Optional[Attr[PhasedArrayXds]] = None

    xradio_version: Optional[Attr[str]] = None
    """ Version of XRADIO used if converted from MSv2. """

    intent: Optional[Attr[str]] = None
    """Identifies the intention of the scan, such as to calibrate or observe a
    target. See :ref:`scan intents` for possible values.
    """
    data_description_id: Optional[Attr[str]] = None
    """
    The id assigned to this combination of spectral window and polarization setup.
    """



def make_quantity(value, units: str) -> dict:
    """
    create a quantity dictionary given value and units
    Parameters
    ----------
    value : numeric or array of numerics
        Quantity value
    units: str
        Quantity units
    Returns
    -------
    dict
    """
    return {"value": value, "units": units, "type": "quantity"}



import toolviper.utils.logger as logger
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import xarray as xr

from .msv2_msv3 import ignore_msv2_cols
from .partition_queries import (
    make_partition_ids_by_ddi_intent,
)
from .subtables import subt_rename_ids, add_pointing_to_partition
from .descr import describe_ms
from ._tables.read import load_generic_table, make_freq_attrs
from ._tables.read_main_table import read_flat_main_table, read_expanded_main_table
from .._utils.partition_attrs import add_partition_attrs
from .._utils.xds_helper import expand_xds, make_coords, optimal_chunking


PartitionKey = Tuple[Any, ...]
VisSetMetaInfo = Dict[str, xr.Dataset]
VisSetPartitions = Dict[PartitionKey, xr.Dataset]


def make_spw_names_by_ddi(ddi_xds: xr.Dataset, spw_xds: xr.Dataset) -> Dict[int, str]:
    spw_ids_by_ddi = ddi_xds.SPECTRAL_WINDOW_ID[ddi_xds.row].values
    spw_names = spw_xds.NAME[spw_ids_by_ddi].values
    return {ddi: spw_names[ddi] for ddi in np.arange(0, len(spw_names))}


def split_intents(intents: str):
    """
    Make a dict with two scan / subscan levels of intents from an
    intent string from the STATE/OBS_MODE of an MS.

    Parameters
    ----------
    intents : str
        intents "OBS_MODE" string from an MS/STATE row

    Returns
    -------
    Dict[str, list]
        per scan intent list of individual subscan intent strings
    """
    sub_sep = "#"
    if sub_sep not in intents:
        sub_sep = "."
        if sub_sep not in intents:
            return intents

    indiv = [intnt for intnt in intents.split(",")]
    scan_subscan_intents = {}
    for ind in indiv:
        scan = ind.split(sub_sep)
        if len(scan) == 1:
            scan = scan[0]
            subscan = ""
        elif len(scan) == 2:
            scan, subscan = scan
        if scan in scan_subscan_intents:
            scan_subscan_intents[scan].append(subscan)
        else:
            scan_subscan_intents[scan] = [subscan]

    return scan_subscan_intents


def make_part_key(
    xds: xr.Dataset,
    partition_scheme: str,
    intent: str = "",
    scan_state: Union[Tuple, None] = None,
) -> PartitionKey:
    """
    Makes the key that a partition (sub)xds will have in the partitions dictionary of a cds.

    Parameters
    ----------
    xds : xr.Dataset
        partition xds with data and attrs
    partition_scheme : str
        one of the schemes supported in the read_ms_*_partitions() functions
    intent : str (Default value = "")
        partition intent
    scan_state : Union[Tuple, None] (Default value = None)
        scan/state ids, required when partition_scheme != 'ddi'

    Returns
    -------
    PartitionKey
        partition key
    """
    spw_id = xds.attrs["partition_ids"]["spw_id"]
    pol_setup_id = xds.attrs["partition_ids"]["pol_setup_id"]

    if partition_scheme == "ddi":
        part_key = (spw_id, pol_setup_id)
    elif partition_scheme == "intent":
        part_key = (spw_id, pol_setup_id, intent)
    elif partition_scheme == "scan":
        scan, _state = scan_state
        part_key = (spw_id, pol_setup_id, scan)
    elif partition_scheme == "scan/subscan":
        scan, state = scan_state
        part_key = (spw_id, pol_setup_id, scan, state)

    return part_key


def read_ms_scan_subscan_partitions(
    infile: str,
    partition_scheme: str,
    expand: bool = False,
    chunks: Union[Tuple[int], List[int], None] = None,
) -> Tuple[VisSetPartitions, Dict[str, xr.Dataset], List[str]]:
    """
    partitions per scan_number/subscans
    (main table column SCAN_NUMBER / STATE_ID)

    Parameters
    ----------
    infile : str
        MS path (main table)
    partition_scheme : str
        this functions can do 'intent', 'scan', and 'scan/subscan'
    expand : bool (Default value = False)
        wether to use (time, baseline) dimensions rather than 1d (row)
        (only relevant when using the read_flat variant of read functions)
    chunk : Union[Tuple[int], List[int], None] (Default value = None)
        Dask chunking as tuple (time, baseline, chan, pol)

    Returns
    -------
    Tuple[VisSetPartitions, Dict[str, xr.Dataset], List[str]]
        a dictionary of partitions, a dict of subtable
        xr.Datasets to use later for metainformation, and a list of the
        subtables already read
    """

    spw_xds = load_generic_table(
        infile,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    )
    ddi_xds = load_generic_table(infile, "DATA_DESCRIPTION")

    if partition_scheme == "intent":
        spw_names_by_ddi = make_spw_names_by_ddi(ddi_xds, spw_xds)
        (
            data_desc_id,
            scan_number,
            state_id,
            distinct_intents,
        ) = make_partition_ids_by_ddi_intent(infile, spw_names_by_ddi)
    else:
        raise ValueError("foo")

    ant_xds = load_generic_table(
        infile, "ANTENNA", rename_ids=subt_rename_ids["ANTENNA"]
    )
    pol_xds = load_generic_table(
        infile, "POLARIZATION", rename_ids=subt_rename_ids["POLARIZATION"]
    )

    # TODO: change to a map?
    partitions = {}
    cnt = 0
    for ddi, scan, state in zip(data_desc_id, scan_number, state_id):
        # msg_chunks = str(chunks[ddi] if type(chunks) == dict else chunks)
        if partition_scheme == "intent":
            intent = distinct_intents[cnt]
            cnt += 1
        else:
            intent = ""

        if partition_scheme == "scan":
            scan_state = (scan, None)
        else:
            scan_state = (scan, state)
        # experimenting, comparing overheads of expanded vs. flat
        expanded = not expand
        if expanded:
            xds, part_ids, attrs = read_expanded_main_table(
                infile, ddi, scan_state=scan_state, ignore_msv2_cols=ignore_msv2_cols
            )
        else:
            xds, part_ids, attrs = read_flat_main_table(
                infile,
                ddi,
                scan_state=scan_state,
                rowidxs=None,
                ignore_msv2_cols=ignore_msv2_cols,
            )
        if len(xds.sizes) == 0:
            continue

        coords = make_coords(xds, ddi, (ant_xds, ddi_xds, spw_xds, pol_xds))
        xds = xds.assign_coords(coords)

        if partition_scheme == "intent":
            scan_subscan_intents = split_intents(intent)
            attrs = dict({"scan_subscan_intents": scan_subscan_intents}, **attrs)
        xds = add_partition_attrs(xds, ddi, ddi_xds, part_ids, attrs)

        # freq dim needs to pull its units/measure info from the SPW subtable
        spw_id = xds.attrs["partition_ids"]["spw_id"]
        xds.freq.attrs.update(make_freq_attrs(spw_xds, spw_id))

        # expand the row dimension out to (time, baseline)
        if not expanded and expand:
            xds = expand_xds(xds)

        part_key = make_part_key(
            xds, partition_scheme, scan_state=scan_state, intent=intent
        )
        partitions[part_key] = xds

    subtables = {
        "antenna": ant_xds,
        "spectral_window": spw_xds,
        "polarization": pol_xds,
    }

    return (
        partitions,
        subtables,
        ["ANTENNA", "SPECTRAL_WINDOW", "POLARTIZATION", "DATA_DESCRIPTION"],
    )


def read_ms_ddi_partitions(
    infile: str,
    expand: bool = False,
    rowmap: Union[dict, None] = None,
    chunks: Union[Tuple[int], List[int], None] = None,
) -> Tuple[VisSetPartitions, Dict[str, xr.Dataset], List[str]]:
    """
    Reads data columns from the main table into partitions defined
    from the DDIs. First looks into the SPECTRAL_WINDOW, POLARIZATION,
    DATA_DESCRIPTION tables to define the partitions.

    Parameters
    ----------
    infile : str
        input MS path
    expand : bool (Default value = False)
        redimension (row)->(time,baseline)
    rowmap : Union[dict, None] (Default value = None)
        to be removed
    chunks : Union[Tuple[int], List[int], None] (Default value = None)
        array data chunk sizes

    Returns
    -------
    Tuple[VisSetPartitions, Dict[str, xr.Dataset], List[str]]
        dictionary of partitions, dict of subtable xr.Datasets to use later
        for metainformation, and a list of the subtables already read

    """
    # we need the antenna, spectral window, polarization, and data description tables
    # to define the (sub)datasets (their dims and coords) and to process the main table
    ant_xds = load_generic_table(
        infile, "ANTENNA", rename_ids=subt_rename_ids["ANTENNA"]
    )
    spw_xds = load_generic_table(
        infile,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    )
    pol_xds = load_generic_table(
        infile, "POLARIZATION", rename_ids=subt_rename_ids["POLARIZATION"]
    )
    ddi_xds = load_generic_table(infile, "DATA_DESCRIPTION")

    # each DATA_DESC_ID (ddi) is a fixed shape that may differ from others
    # form a list of ddis to process, each will be placed it in its own xarray dataset and partition
    ddis = np.arange(ddi_xds.row.shape[0]) if rowmap is None else list(rowmap.keys())

    # figure out the chunking for each DDI, either one fixed shape or an auto-computed one
    if type(chunks) is not tuple:
        mshape = describe_ms(infile, mode="flat", rowmap=rowmap)
        chunks = dict(
            [
                (
                    ddi,
                    optimal_chunking(
                        didxs=chunks, chunk_size="auto", data_shape=mshape[ddi]
                    ),
                )
                for ddi in mshape
            ]
        )

    partitions = {}
    ####################################################################
    # process each selected DDI from the input MS, assume a fixed shape within the ddi (should always be true)
    for ddi in ddis:
        rowidxs = None if rowmap is None else rowmap[ddi][0]
        chanidxs = None if rowmap is None else rowmap[ddi][1]
        if ((rowidxs is not None) and (len(rowidxs) == 0)) or (
            (chanidxs is not None) and (len(chanidxs) == 0)
        ):
            continue
        logger.debug(
            "reading DDI %i with chunking %s..."
            % (ddi, str(chunks[ddi] if type(chunks) is dict else chunks))
        )

        # experimenting, comparing overheads of expanded vs. flat
        expanded = not expand
        if expanded:
            xds, part_ids, attrs = read_expanded_main_table(
                infile, ddi, ignore_msv2_cols=ignore_msv2_cols
            )
        else:
            xds, part_ids, attrs = read_flat_main_table(
                infile,
                ddi,
                rowidxs=rowidxs,
                chunks=chunks[ddi] if type(chunks) is dict else chunks,
                ignore_msv2_cols=ignore_msv2_cols,
            )
        if len(xds.sizes) == 0:
            continue

        coords = make_coords(xds, ddi, (ant_xds, ddi_xds, spw_xds, pol_xds))
        xds = xds.assign_coords(coords)

        xds = add_partition_attrs(xds, ddi, ddi_xds, part_ids, attrs)

        # freq dim needs to pull its units/measure info from the SPW subtable
        spw_id = xds.attrs["partition_ids"]["spw_id"]
        xds.freq.attrs.update(make_freq_attrs(spw_xds, spw_id))

        # filter by channel selection
        if (chanidxs is not None) and (len(chanidxs) < len(xds.chan)):
            xds = xds.isel(chan=chanidxs)
            spw_xds["CHAN_FREQ"][
                ddi_xds.SPECTRAL_WINDOW_ID.values[ddi], : len(chanidxs)
            ] = spw_xds.CHAN_FREQ[ddi_xds.SPECTRAL_WINDOW_ID.values[ddi], chanidxs]

        # expand the row dimension out to (time, baseline)
        if not expanded and expand:
            xds = expand_xds(xds)

        part_key = make_part_key(xds, partition_scheme="ddi")
        partitions[part_key] = xds

    subtables = {
        "antenna": ant_xds,
        "spectral_window": spw_xds,
        "polarization": pol_xds,
        "data_description": ddi_xds,
    }

    return (
        partitions,
        subtables,
        ["ANTENNA", "SPECTRAL_WINDOW", "POLARTIZATION", "DATA_DESCRIPTION"],
    )


def finalize_partitions(
    parts: Dict[str, xr.Dataset], subts: Dict[str, xr.Dataset]
) -> Dict[str, xr.Dataset]:
    """
    Once the partition datasets and the metainfo/subtable datasets
    have been read, add to the partitions:
    - pointing variables from the pointing subtable

    Parameters
    ----------
    parts : Dict[str, xr.Dataset]
        partitions as xarray datasets, as read from an MS main table
    subts : Dict[str, xr.Dataset]
        subtables of an MS read as xarray datasets

    Returns
    -------
    Dict[str, xr.Dataset]
        partitions with additions taken from subtables
    """
    if "pointing" in subts:
        pointing = subts["pointing"]
        final = {
            key: add_pointing_to_partition(xds, pointing)
            for (key, xds) in parts.items()
        }
    else:
        final = parts

    return final



from .dataclass import (
    xarray_dataclass_to_array_schema,
    xarray_dataclass_to_dataset_schema,
    xarray_dataclass_to_dict_schema,
)
from .bases import (
    xarray_dataarray_schema,
    xarray_dataset_schema,
    dict_schema,
)
from .check import (
    SchemaIssue,
    SchemaIssues,
    check_array,
    check_dataset,
    check_dict,
    schema_checked,
)

__all__ = [
    "xarray_dataclass_to_array_schema",
    "xarray_dataclass_to_dataset_schema",
    "xarray_dataclass_to_dict_schema",
    "AsDataArray",
    "AsDataset",
    "AsDict",
    "xarray_dataarray_schema",
    "xarray_dataset_schema",
    "SchemaIssue",
    "SchemaIssues",
    "check_array",
    "check_dataset",
    "check_dict",
    "schema_checked",
]



import numpy as np
import xarray as xr
import os
from .common import _np_types, _top_level_sub_xds


def _write_zarr(xds: xr.Dataset, zarr_store: str):
    xds_copy = xds.copy(deep=True)
    xds_copy, xds_dict = _encode(xds_copy)
    z_obj = xds_copy.to_zarr(store=zarr_store, compute=True)
    _write_sub_xdses(zarr_store, xds_dict, _top_level_sub_xds)


def _encode(xds: xr.Dataset):
    # encode attrs
    xds.attrs, xds_dict = _encode_dict(xds.attrs)
    return xds, xds_dict


def _encode_dict(my_dict: dict, top_key="") -> tuple:
    xds_dict = {}
    del_keys = []
    for k, v in my_dict.items():
        if isinstance(v, dict):
            z = os.sep.join([top_key, k]) if top_key else k
            my_dict[k], ret_xds_dict = _encode_dict(v, z)
            if ret_xds_dict:
                xds_dict[k] = ret_xds_dict
        elif isinstance(v, np.ndarray):
            my_dict[k] = {}
            my_dict[k]["_type"] = "numpy.ndarray"
            my_dict[k]["_value"] = v.tolist()
            my_dict[k]["_dtype"] = str(v.dtype)
        elif isinstance(v, xr.Dataset):
            xds_dict[k] = v.copy(deep=True)
            del_keys.append(k)
    for k in del_keys:
        del my_dict[k]
    return my_dict, xds_dict


def _write_sub_xdses(zarr_store: str, xds_dict: dict, path: str):
    for k, v in xds_dict.items():
        my_path = f"{path}_{k}" if path else f"{k}"
        if isinstance(v, dict):
            _write_sub_xdses(zarr_store, xds_dict[k], my_path)
        elif isinstance(v, xr.Dataset):
            zs = os.sep.join([zarr_store, my_path])
            z_obj = v.to_zarr(store=zs, compute=True)



from typing import get_type_hints, get_args
from .typing import get_dims, get_types, get_role, Role, get_annotated, is_optional

import typing
import inspect
import ast
import dataclasses
import numpy
import itertools
import textwrap

from xradio.schema.metamodel import *


def extract_field_docstrings(klass):
    """
    Extracts docstrings of all fields from a class

    :param klass: Class to extract docstrings from
    :returns: Dictionary of attributes to clean docstrings
    """

    # Parse class body
    try:
        src = inspect.getsource(klass)
    except OSError:
        return {}
    module = ast.parse(textwrap.dedent(src))

    # Expect module containing a class definition
    if not isinstance(module, ast.Module) or len(module.body) != 1:
        raise valueError(f"Expected parser to generate ast.Module, got {module}!")
    cls = module.body[0]
    if not isinstance(cls, ast.ClassDef):
        raise valueError(f"Expected a class definition, got {ast.dump(cls)}!")

    # Go through body, collect dostrings
    docstrings = {}
    for i, assign in enumerate(cls.body):
        # Handle both annotated and unannotated case
        if isinstance(assign, ast.AnnAssign):
            if not isinstance(assign.target, ast.Name):
                warnings.warn(f"Expected name in assignment {ast.dump(assign)}!")
                continue
            names = [assign.target.id]
        elif isinstance(assign, ast.Assign):
            if not all(isinstance(name, ast.Name) for name in assign.targets):
                warnings.warn(f"Expected names in assignment {ast.dump(assign)}!")
                continue
            names = [name.id for name in assign.targets]
        else:
            continue

        # Get docstring from next statement
        if (
            i + 1 < len(cls.body)
            and isinstance(cls.body[i + 1], ast.Expr)
            and isinstance(cls.body[i + 1].value, ast.Constant)
            and isinstance(cls.body[i + 1].value.value, str)
        ):
            for name in names:
                docstrings[name] = inspect.cleandoc(cls.body[i + 1].value.value)
        else:
            for name in names:
                docstrings[name] = None

    return docstrings


def _check_invalid_dims(
    dims: list[list[str]], all_coord_names: list[str], klass_name: str, field_name: str
):
    """
    Check dimension possibilities for undefined coordinates
    """

    # Filter out dimension possibilities with undefined coordinates
    valid_dims = [ds for ds in dims if set(ds).issubset(all_coord_names)]
    # print(f"{klass_name}.{field_name}", valid_dims, dims, all_coord_names)

    # Raise an exception if this makes the dimension set impossible
    if dims and not valid_dims:
        required_dims = sorted(map(lambda ds: set(ds) - all_coord_names, dims), key=len)
        raise ValueError(
            f"In '{klass_name}', field '{field_name}' has"
            f" undefined coordinates, consider defining {required_dims}!"
        )
    return valid_dims


def extract_xarray_dataclass(klass, allow_undefined_coords: bool = False):
    """
    Go through dataclass fields and interpret them according to xarray-dataclass

    Returns a tuple of coordinates, data variables and attributes

    :param allow_undefined_coords: Allow data variables with dimensions
      that do not have associated coordinates (e.g. for data arrays).
    """

    field_docstrings = extract_field_docstrings(klass)

    # Collect type hints, identify coordinates
    type_hints = get_type_hints(klass, include_extras=True)
    if allow_undefined_coords:

        def check_invalid_dims(dims, field_name):
            return dims

    else:
        all_coord_names = {
            field.name
            for field in dataclasses.fields(klass)
            if get_role(type_hints[field.name]) == Role.COORD
        }

        def check_invalid_dims(dims, field_name):
            return _check_invalid_dims(
                dims, all_coord_names, klass.__name__, field_name
            )

    # Go through attributes, collecting coordinates, data variables and
    # attributes
    coordinates = []
    data_vars = []
    attributes = []
    for field in dataclasses.fields(klass):
        # Get field "role" (coordinate / data variable / attribute) from its
        # type hint
        typ = type_hints[field.name]
        role = get_role(typ)

        # Is it an attribute?
        if role == Role.ATTR:
            attributes.append(
                AttrSchemaRef(
                    name=field.name,
                    typ=get_annotated(typ),
                    optional=is_optional(typ),
                    default=field.default,
                    docstring=field_docstrings.get(field.name),
                )
            )
            continue

        # Only other allowed option is some kind of data array
        if role == Role.COORD:
            is_coord = True
        elif role == Role.DATA:
            is_coord = False
        else:
            raise ValueError(
                f"Expected field '{field.name}' in '{klass.__name__}' "
                "to be annotated with either Coord, Data or Attr!"
            )

        # Defined using a dataclass, i.e. Coordof/Dataof?
        dataclass = typing.get_args(get_annotated(typ))[0]
        if dataclasses.is_dataclass(dataclass):
            # Recursively get array schema for data class
            arr_schema = xarray_dataclass_to_array_schema(dataclass)
            arr_schema_fields = {
                f.name: getattr(arr_schema, f.name)
                for f in dataclasses.fields(ArraySchema)
            }

            # Check for undefined coordinates
            arr_schema_fields["dimensions"] = check_invalid_dims(
                arr_schema_fields["dimensions"], field.name
            )

            # Repackage as reference
            schema_ref = ArraySchemaRef(
                name=field.name,
                optional=is_optional(typ),
                default=field.default,
                docstring=field_docstrings.get(field.name),
                **arr_schema_fields,
            )

        else:
            # Get dimensions and dtypes
            dims = get_dims(typ)
            types = get_types(typ)

            # Is types a (single) dataclass?
            if len(types) == 1 and dataclasses.is_dataclass(types[0]):
                # Recursively get array schema for data class
                arr_schema = xarray_dataclass_to_array_schema(types[0])

                # Prepend dimensions to array schema
                combined_dimensions = [
                    dims1 + dims2
                    for dims1, dims2 in itertools.product(dims, arr_schema.dimensions)
                ]

                # Repackage as reference
                arr_schema_fields = {
                    f.name: getattr(arr_schema, f.name)
                    for f in dataclasses.fields(ArraySchema)
                }

                arr_schema_fields["dimensions"] = check_invalid_dims(
                    combined_dimensions, field.name
                )
                schema_ref = ArraySchemaRef(
                    name=field.name,
                    optional=is_optional(typ),
                    default=field.default,
                    docstring=field_docstrings.get(field.name),
                    **arr_schema_fields,
                )
            else:
                # Assume that it's an "inline" declaration using "Coord"/"Data"
                schema_ref = ArraySchemaRef(
                    name=field.name,
                    optional=is_optional(typ),
                    default=field.default,
                    docstring=field_docstrings.get(field.name),
                    schema_name=None,
                    dimensions=check_invalid_dims(dims, field.name),
                    dtypes=[numpy.dtype(typ) for typ in types],
                    coordinates=[],
                    attributes=[],
                    class_docstring=None,
                    data_docstring=None,
                )

        if is_coord:
            # Make sure that it is valid to use as a coordinate - i.e. we don't
            # have "recursive" (?!) coordinate definitions
            if not schema_ref.is_coord():
                raise ValueError(
                    f"In '{klass.__name__}', field '{field.name}':"
                    f" {schema_ref.schema_name} has coordinates, and"
                    " therefore can't be used as coordinate array itself!"
                )

            coordinates.append(schema_ref)
        else:
            data_vars.append(schema_ref)

    return coordinates, data_vars, attributes


def xarray_dataclass_to_array_schema(klass):
    """
    Convert an xarray-dataclass schema dataclass to an ArraySchema

    This should work on any class that we would derive from AsDataArray, or
    refer to using CoordOf or DataOf
    """

    # Cached?
    if hasattr(klass, "__xradio_array_schema"):
        return klass.__xradio_array_schema

    # Extract from data class
    coordinates, data_vars, attributes = extract_xarray_dataclass(klass, True)

    # For a dataclass there must be exactly one data variable
    if not data_vars:
        raise ValueError(
            f"Found no data declaration in (supposed) data array class {klass.__name__}!"
        )
    if len(data_vars) > 1:
        raise ValueError(
            f"Found multiple data variables ({', '.join(v.name for v in data_vars)})"
            f" in supposed data array class {klass.__name__}!"
        )

    # Check that data variable is named "data". This is important for this to
    # match up with parameters to xarray.DataArray() later (see bases.AsArray)
    if data_vars[0].name != "data":
        raise ValueError(
            f"Data variable in data array class {klass.__name__} "
            f'must be called "data", not {data_vars[0].name}!'
        )

    # Make class
    schema = ArraySchema(
        schema_name=f"{klass.__module__}.{klass.__qualname__}",
        dimensions=data_vars[0].dimensions,
        dtypes=data_vars[0].dtypes,
        coordinates=coordinates,
        attributes=attributes,
        class_docstring=inspect.cleandoc(klass.__doc__),
        data_docstring=data_vars[0].docstring,
    )
    klass.__xradio_array_schema = schema
    return schema


def xarray_dataclass_to_dataset_schema(klass):
    """
    Convert an xarray-dataclass schema dataclass to an ArraySchema

    This should work on any class that we would derive from AsDataArray, or
    refer to using CoordOf or DataOf
    """

    # Cached?
    if hasattr(klass, "__xradio_dataset_schema"):
        return klass.__xradio_dataset_schema

    # Extract from data class
    coordinates, data_vars, attributes = extract_xarray_dataclass(klass)

    # Collect all possible dimensions (use dict instead of set so we retain
    # ordering)
    all_dvars = coordinates + data_vars
    all_dimensions = dict()
    for dvar in all_dvars:
        for dims in dvar.dimensions:
            all_dimensions.update(dict.fromkeys(dims))

    # Identify dimensions that are always required, because it is mentioned in
    # all possible dimensionalities on a non-optional data variable
    req_dimensions = set()
    for dvar in all_dvars:
        if not dvar.optional:
            req_dimensions |= dvar.required_dimensions()

    # Build powerset of all options
    # (Theoretically we could try to be fancy here and filter out options
    #  that "make no sense". E.g. if there are some dimensions A and B
    #  that are always used together, but never appear individually,
    #  we should not list A and B separately as options. The test would
    #  be to check whether for every dimensionality option there exists
    #  a mapping of data variables to allowed dimensions [possibly empty
    #  if the data variable is optional], such that the superset of all
    #  dimensions matches. Likely overkill for now.)
    opt_dimensions = all_dimensions.keys() - req_dimensions
    dimensions = [req_dimensions]
    for dim in opt_dimensions:
        dimensions += [dims | set([dim]) for dims in dimensions]

    # Reorder consistently / convert to lists
    dimensions = [[dim for dim in all_dimensions if dim in dims] for dims in dimensions]

    # Make class
    schema = DatasetSchema(
        schema_name=f"{klass.__module__}.{klass.__qualname__}",
        dimensions=dimensions,
        coordinates=coordinates,
        data_vars=data_vars,
        attributes=attributes,
        class_docstring=inspect.cleandoc(klass.__doc__),
    )
    klass.__xradio_dataset_schema = schema
    return schema


def xarray_dataclass_to_dict_schema(klass):
    """
    Convert an xarray-dataclass style schema dataclass to an DictSchema

    This should work on any class annotated with :py:func:`~xradio.schema.bases.dict_schema`
    """

    # Cached?
    if hasattr(klass, "__xradio_dict_schema"):
        return klass.__xradio_dict_schema

    # Get docstrings and type hints
    field_docstrings = extract_field_docstrings(klass)
    type_hints = get_type_hints(klass, include_extras=True)
    attributes = []
    for field in dataclasses.fields(klass):
        typ = type_hints[field.name]

        # Handle optional value: Strip "None" from the types
        optional = is_optional(typ)
        if optional:
            typs = [typ for typ in get_args(typ) if typ is not None.__class__]
            if len(typs) == 1:
                typ = typs[0]
            else:
                typ = typing.Union.__getitem__[tuple(typs)]

        attributes.append(
            AttrSchemaRef(
                name=field.name,
                typ=typ,
                optional=optional,
                default=field.default,
                docstring=field_docstrings.get(field.name),
            )
        )

    # Return
    schema = DictSchema(
        schema_name=f"{klass.__module__}.{klass.__qualname__}",
        attributes=attributes,
        class_docstring=inspect.cleandoc(klass.__doc__),
    )
    klass.__xradio_dict_schema = schema
    return schema



from typing import Tuple

import numpy as np

from casacore import tables
from ....._utils.common import get_pad_value


def load_col_chunk(
    tb_tool: tables.table,
    col: str,
    cshape: Tuple[int],
    tidxs: np.ndarray,
    bidxs: np.ndarray,
    didxs: np.ndarray,
    d1: Tuple[int, int],
    d2: Tuple[int, int],
) -> np.ndarray:
    """
    Loads a slice of a col (using casacore getcol(slice))

    Parameters
    ----------
    tb_tool : tables.table
        a table/TaQL query open and being used to load columns
    col : str
        colum to load
    cshape : Tuple[int]
        shape of the resulting col data chunk
    tidxs : np.ndarray
        time axis indices
    bidxs : np.ndarray
        baseline axis indices
    didxs : np.ndarray
        effective) data indices, excluding missing baselines
    d1 : Tuple[int, int]
        indices to load on dimension 1 (None=all, pols or chans)
    d2 : Tuple[int, int]
        indices to load on dimension 2 (None=all, pols)

    Returns
    -------
    np.ndarray
        data array loaded directly with casacore getcol/getcolslice
    """

    if (len(cshape) == 2) or (col == "UVW"):
        # all the scalars and UVW
        data = np.array(tb_tool.getcol(col, 0, -1))
    elif len(cshape) == 3:
        # WEIGHT, SIGMA
        data = tb_tool.getcolslice(col, d1[0], d1[1], [], 0, -1)
    elif len(cshape) == 4:
        # DATA, FLAG, WEIGHT_SPECTRUM / SIGMA_SPECTRUM
        data = tb_tool.getcolslice(col, (d1[0], d2[0]), (d1[1], d2[1]), [], 0, -1)

    # full data is the maximum of the data shape and chunk shape dimensions
    fill_value = get_pad_value(data.dtype)
    chunk = np.full(cshape, np.fill_value, dtype=data.dtype)
    if len(didxs) > 0:
        chunk[tidxs[didxs], bidxs[didxs]] = data[didxs]

    return chunk



import os
import toolviper.utils.logger as logger
from typing import List, Tuple, Union

from xradio.measurement_set._utils._utils.cds import CASAVisSet
from xradio.measurement_set._utils._msv2.partitions import (
    finalize_partitions,
    read_ms_ddi_partitions,
    read_ms_scan_subscan_partitions,
)
from xradio.measurement_set._utils._msv2.subtables import read_ms_subtables
from xradio.measurement_set._utils._utils.xds_helper import vis_xds_packager_cds


def read_ms(
    infile: str,
    subtables: bool = True,
    asdm_subtables: bool = False,
    partition_scheme: str = "intent",
    chunks: Union[Tuple[int], List[int]] = None,
    expand: bool = False,
    **kwargs: str,
) -> CASAVisSet:
    """
    Read a MeasurementSet (MSv2 format) into a next generation CASA
    dataset (visibilities dataset as a set of Xarray datasets).

    The MS is partitioned into multiple sub- Xarray datasets (where the data variables are read as
    Dask delayed arrays).
    The MS is partitioned by DDI, which guarantees a fixed data shape per partition (in terms of channels
    and polarizations) and, subject to experimentation, by scan and subscan. This results in multiple
    partitions as xarray datasets (xds) contained within a main xds (mxds).

    Parameters
    ----------
    infile : str
        Input MS filename
    subtables : bool (Default value = True)
        Also read and include subtables along with main table selection. Default False will
        omit subtables (faster)
    asdm_subtables : bool (Default value = False)
        in addition to MeasurementSet subtables (if enabled), also read extension
        subtables named "ASDM_*"
    partition_scheme : str (Default value = "intent")
        experimenting) Whether to partition sub-xds datasets by scan/subscan
        (in addition to DDI), or other alternative partitioning schemes. Accepted values: 'scan/subscan',
        'scan', 'ddi', 'intent'. Default: 'intent'
    chunks : Union[Tuple[int], List[int]] (Default value = None)
        Can be used to set a specific chunk shape (with a tuple of ints), or to control the
        optimization used for automatic chunking (with a list of ints). A tuple of ints in the form of (row,
        chan, pol) will use a fixed chunk shape. A list or numpy array of ints in the form of [idx1, etc]
        will trigger auto-chunking optimized for the given indices, with row=0, chan=1, pol=2. Default None
        uses auto-chunking with a best fit across all dimensions (probably sub-optimal for most cases).
    expand : bool (Default value = False)
        to be removed) Whether or not to return the original flat row structure of the MS (False)
        or expand the rows to time x baseline dimensions (True). Expanding the rows allows for easier indexing
        and parallelization across time and baseline dimensions, at the cost of some conversion time.
    **kwargs: str :


    Returns
    -------
    CASAVisSet
        Main xarray dataset of datasets for this visibility dataset
    """

    infile = os.path.expanduser(infile)
    if not os.path.isdir(infile):
        raise ValueError(f"invalid input filename to read_ms {infile}")

    # Several alternatives to experiment for now
    part_descr = {
        "intent": "scan/subscan intent + DDI",
        "ddi": "DDI",
        "scan": "scan + DDI",
        "scan/subscan": "scan + subscan + DDI",
    }

    if partition_scheme not in part_descr:
        raise ValueError(f"Invalid partition_scheme: {partition_scheme}")

    logger.info(
        f"Reading {infile} as MSv2 and applying partitioning by {part_descr[partition_scheme]}"
    )

    if partition_scheme == "ddi":
        logger.info(f"Reading {infile} as MSv2 and applying DDI partitioning")
        # get the indices of the ms selection (if any)
        # rowmap = ms_selection(infile, verbose=verbose, **kwargs)
        rowmap = None
        parts, subts, done_subts = read_ms_ddi_partitions(
            infile, expand, rowmap, chunks
        )
    else:
        parts, subts, done_subts = read_ms_scan_subscan_partitions(
            infile, partition_scheme, expand, chunks
        )

    if subtables:
        subts.update(read_ms_subtables(infile, done_subts, asdm_subtables))

    parts = finalize_partitions(parts, subts)

    # build the visibilities container (metainfo + partitions) to return
    cds = vis_xds_packager_cds(subts, parts, "read_ms")
    return cds



import xarray
import inspect
from . import dataclass, check, metamodel, typing
import numpy
import dataclasses


def _guess_dtype(obj: typing.Any):
    try:
        return _guess_dtype(next(iter(obj)))
    except TypeError:
        return numpy.dtype(type(obj))


def _set_parameter(
    val: typing.Any, args: dict, schema: typing.Union["AttrSchemaRef", "ArraySchemaRef"]
):
    """
    Extract given entry from parameters - while taking care that the
    parameter value might have been set either before or after, and that
    defaults might apply.

    :param val: Value from xarray-constructor style ("data_vars"/"coords")
    :param args: Bound arguments to constructor (positional or named)
    :param schema: Schema of argument (either attribute or array)
    :returns: Updated value
    """

    # If value appears in named parameters, overwrite
    if args.get(schema.name) is not None:
        if val is not None:
            raise ValueError(
                f"Parameter {schema.name} was passed twice ({val} vs {args[schema.name]})!"
            )
        val = args[schema.name]

    # Otherwise apply defaults *if* it doesn't exist already or deactivate
    # (typically because we are construting from a dataset/data array)
    if val is None and schema.default is not dataclasses.MISSING:
        default = schema.default
        if default is not None:
            val = default

    return val


def _np_convert(val: typing.Any, schema: metamodel.ArraySchemaRef):
    """
    Convert value to numpy, if appropriate

    This attempts to catch "early" conversions that we can do more
    appropriately than xarray because we have more information from the schema.
    Specifically, if it's a type where the dtype to choose is somewhat
    ambiguous, we can use this chance to "bias" it towards an allowed one.

    :param val: Received value
    :param schema: Execpted array schema
    :returns: Possibly converted value
    """

    # Array schema refs that are not yet a numpy or xarray data type?
    if isinstance(val, list) or isinstance(val, tuple) and isinstance(val[1], list):
        # Check whether we can "guess" the dtype from the object
        dtype = None
        if len(schema.dtypes) > 1:
            guessed = _guess_dtype(val)
            for dt in schema.dtypes:
                # Actually look for closest in precision etc?
                if dt == guessed:
                    dtype = dt
                    break

        # Otherwise just use the first one
        if dtype is None:
            dtype = schema.dtypes[0]

        # Attempt conversation
        try:
            if isinstance(val, list):
                val = numpy.array(val, dtype=dtype)
            else:
                val = tuple([val[0], numpy.array(val[1], dtype=dtype), *val[2:]])

        except TypeError:
            pass

    return val


def _dataarray_new(
    cls,
    data=None,
    *args,
    coords=None,
    dims=None,
    name=None,
    attrs=None,
    indexes=None,
    **kwargs,
):
    # Convert parameters
    if coords is not None and isinstance(coords, list):
        coords = dict(coords)
    if coords is None:
        coords = {}
    if attrs is None:
        attrs = {}

    # Get signature of __init__, map parameters and apply defaults. This
    # will raise an exception if there are any extra parameters.
    sig = inspect.Signature.from_callable(cls.__init__)
    sig = sig.replace(parameters=[v for k, v in sig.parameters.items() if k != "self"])
    mapping = sig.bind_partial(data, *args, **kwargs)

    # Check whether we have a "data" argument now. This happens if we pass
    # it as a positional argument.
    if mapping.arguments.get("data") is not None:
        data = mapping.arguments["data"]

    # No dims specified? Select one matching the data dimensionality from
    # the schema
    schema = dataclass.xarray_dataclass_to_array_schema(cls)
    data = _np_convert(data, schema)
    for schema_dims in schema.dimensions:
        if len(schema_dims) == len(data.shape):
            dims = schema_dims
            break

    # If we are constructing from a data array / variable, take over attributes
    if isinstance(data, (xarray.DataArray, xarray.Variable)):
        for attr, attr_val in data.attrs.items():
            # Explicit parameters take precedence though
            if attr not in attrs:
                attrs[attr] = attr_val

    # Get any coordinates or attributes and add them to the appropriate lists
    for coord in schema.coordinates:
        val = _np_convert(
            _set_parameter(coords.get(coord.name), mapping.arguments, coord), coord
        )

        # Default to simple range of specified dtype if part of dimensions
        # (that's roughly the behaviour of the xarray constructor as well)
        if val is None and dims is not None:
            dim_ix = dims.index(coord.name)
            if dim_ix is not None and dim_ix < len(data.shape):
                dtype = coord.dtypes[0]
                val = numpy.arange(data.shape[dim_ix], dtype=dtype)

        if val is not None:
            coords[coord.name] = val
    for attr in schema.attributes:
        val = _set_parameter(attrs.get(attr.name), mapping.arguments, attr)
        if val is not None:
            attrs[attr.name] = val

    # Redirect to xradio.DataArray constructor
    instance = xarray.DataArray(data, coords, dims, name, attrs, indexes)

    # Perform schema check
    check.check_array(instance, schema).expect()
    return instance


def xarray_dataarray_schema(cls):
    """Decorator for classes representing :py:class:`xarray.DataArray` schemas.
    The annotated class should exactly contain:

    * one field called "``data``" annotated with :py:data:`~typing.Data`
      to indicate the array type
    * fields annotated with :py:data:`~typing.Coord` to indicate mappings of
      dimensions to coordinates (coordinates directly associated with dimensions
      should have the same name as the dimension)
    * fields annotated with :py:data:`~typing.Attr` to declare attributes

    Decorated schema classes can be used with
    :py:func:`~xradio.schema.check.check_array` for checking
    :py:class:`xarray.DataArray` objects against the schema. Furthermore, the
    class constructor will be overwritten to generate schema-confirming
    :py:class:`xarray.DataArray` objects.

    For example::

        from xradio.schema import xarray_dataarray_schema
        from xradio.schema.typing import Data, Coord, Attr
        from typing import Optional, Literal
        import dataclasses

        Coo = Literal["coo"]

        @xarray_dataarray_schema
        class TestArray:
            data: Data[Coo, complex]
            coo: Coord[Coo, float]
            attr1: Attr[str]
            attr2: Attr[int] = 123
            attr3: Optional[Attr[int]] = None

    This data class represents a one-dimensional :py:class:`xarray.DataArray`
    with complex data, a ``float`` coordinate and three attributes. Instances of
    this class cannot actually be constructed, instead you will get an appropriate
    :py:class:`xarray.DataArray` object::

        >>> TestArray(data=[1,2,3], attr1="foo")
        <xarray.DataArray (coo: 3)>
        array([1.+0.j, 2.+0.j, 3.+0.j])
        Coordinates:
          * coo      (coo) float64 0.0 1.0 2.0
        Attributes:
            attr1:    foo
            attr2:    123

    Note that:

    * The constructor uses the annotations to identify the role of every parameter
    * The data was automatically converted into a :py:class:`numpy.ndarray`
    * As there was no coordinate given, it was automatically filled with an
      enumeration of the type specified in the annotation
    * Default attribute values were assigned. A value of `None` is interpreted
      as the value attribute being missing.
    * For the returned :py:class:`~xarray.DataArray` object ``data``, ``coo``,
      ``attr1`` and ``attr2`` can be accessed as if they were members. This works
      as long as the names don't collide with :py:class:`~xarray.DataArray` members.

    Positional parameters are also supported, and ``coords`` and ``attrs`` passed as
    keyword arguments can supply additional coordinates and attributes::

        >>> TestArray([1,2,3], [3,4,5], 'bar', coords={'coo_new': ('coo', [3,2,1])}, attrs={'xattr': 'baz'})
        <xarray.DataArray (coo: 3)>
        array([1.+0.j, 2.+0.j, 3.+0.j])
        Coordinates:
            coo_new  (coo) int64 3 2 1
          * coo      (coo) float64 3.0 4.0 5.0
        Attributes:
            xattr:    baz
            attr1:    bar
            attr2:    123

    """

    # Make into a dataclass (might not even be needed at some point?)
    cls = dataclasses.dataclass(cls, init=True, repr=False, eq=False, frozen=True)

    # Make schema
    cls.__xradio_array_schema = dataclass.xarray_dataclass_to_array_schema(cls)

    # Replace __new__
    cls.__new__ = _dataarray_new

    return cls


def is_dataarray_schema(val: typing.Any):
    return type(val) == type and hasattr(val, "__xradio_array_schema")


class AsDataArray:
    __new__ = _dataarray_new


def _dataset_new(cls, *args, data_vars=None, coords=None, attrs=None, **kwargs):
    # Get standard xarray parameters (data_vars, coords, attrs)
    # Note that we only support these as keyword arguments for now
    if data_vars is None:
        data_vars = {}
    if coords is None:
        coords = {}
    if attrs is None:
        attrs = {}

    # Get signature of __init__, map parameters and apply defaults. This
    # will raise an exception if there are any extra parameters.
    sig = inspect.Signature.from_callable(cls.__init__)
    sig = sig.replace(parameters=[v for k, v in sig.parameters.items() if k != "self"])
    mapping = sig.bind_partial(*args, **kwargs)

    # Now get schema for this class and identify which of the parameters
    # where meant to be data variables, coordinates and attributes
    # respectively. Note that we interpret "None" as "missing" here, so
    # setting an attribute to `None` will require passing them as
    # attrs.
    schema = dataclass.xarray_dataclass_to_dataset_schema(cls)
    for coord in schema.coordinates:
        val = _np_convert(
            _set_parameter(coords.get(coord.name), mapping.arguments, coord), coord
        )
        # Determine dimensions / convert to Variable
        if (
            val is not None
            and not isinstance(val, xarray.DataArray)
            and not isinstance(val, xarray.Variable)
            and not isinstance(val, tuple)
        ):
            default_attrs = {
                attr.name: attr.default
                for attr in coord.attributes
                if attr.default is not None
            }
            for dims in coord.dimensions:
                if len(dims) == len(val.shape):
                    val = xarray.Variable(dims, val, default_attrs)
                    break
        if val is not None:
            coords[coord.name] = val
    for data_var in schema.data_vars:
        val = _set_parameter(data_vars.get(data_var.name), mapping.arguments, data_var)

        # Determine dimensions / convert to Variable
        dims = None
        if val is None:
            dims = None
        elif isinstance(val, xarray.Variable):
            dims = val.dims
        elif isinstance(val, xarray.DataArray):
            val = val.variable
            dims = val.dims
        elif isinstance(val, tuple):
            val = xarray.Variable(*val)
            dims = val.dims
        else:
            # We are dealing with a plain value. Try to convert it to numpy first
            val = _np_convert(val, data_var)

            # Then identify dimensions by matching against dimensionality
            dims = None
            for ds in data_var.dimensions:
                if len(ds) == len(val.shape):
                    dims = ds
                    break
            if dims is None:
                options = ["[" + dims.join(",") + "]" for dims in data_var.dimensions]
                raise ValueError(
                    f"Data variable {data_var.name} shape has {len(dims)} dimensions,"
                    f" expected {' or '.join(options)}!"
                )

            # Get default attributes
            default_attrs = {
                attr.name: attr.default
                for attr in data_var.attributes
                if attr.default is not dataclasses.MISSING
            }

            # Replace by variable
            val = xarray.Variable(dims, val, default_attrs)

        # Default coordinates used by this data variable to numpy arange. We
        # can only do this now because we need an example to determine the
        # intended size of the coordinate
        if dims is not None:
            for coord in schema.coordinates:
                if coord.name in dims and coords.get(coord.name) is None:
                    dim_ix = dims.index(coord.name)
                    if dim_ix is not None and dim_ix < len(val.shape):
                        dtype = coord.dtypes[0]
                        if numpy.issubdtype(dtype, numpy.number):
                            coords[coord.name] = numpy.arange(
                                val.shape[dim_ix], dtype=dtype
                            )

        if val is not None:
            data_vars[data_var.name] = val

    for attr in schema.attributes:
        val = _set_parameter(attrs.get(attr.name), mapping.arguments, attr)
        if val is not None:
            attrs[attr.name] = val

    # Redirect to xradio.Dataset constructor
    instance = xarray.Dataset(data_vars, coords, attrs)

    # Finally check schema
    check.check_dataset(instance, schema).expect()

    return instance


def xarray_dataset_schema(cls):
    """Decorator for classes representing :py:class:`xarray.Dataset` schemas.
    The annotated class should exactly contain:

    * fields annotated with :py:data:`~typing.Coord` to indicate mappings of
      dimensions to coordinates (coordinates directly associated with dimensions
      should have the same name as the dimension)
    * fields annotated with :py:data:`~typing.Data`
      to indicate data variables
    * fields annotated with :py:data:`~typing.Attr` to declare attributes

    Decorated schema classes can be used with
    :py:func:`~xradio.schema.check.check_dataset` for checking
    :py:class:`xarray.Dataset` objects against the schema. Furthermore, the
    class constructor will be overwritten to generate schema-confirming
    :py:class:`xarray.Dataset` objects.
    """

    # Make into a dataclass (might not even be needed at some point?)
    cls = dataclasses.dataclass(cls, init=True, repr=False, eq=False, frozen=True)

    # Make schema
    cls.__xradio_dataset_schema = dataclass.xarray_dataclass_to_dataset_schema(cls)

    # Replace __new__
    cls.__new__ = _dataset_new

    return cls


def is_dataset_schema(val: typing.Any):
    return type(val) == type and hasattr(val, "__xradio_dataset_schema")


class AsDataset:
    """Mix-in class to indicate dataset data classes

    Deprecated - use decorator :py:func:`xarray_dataset_schema` instead
    """

    __new__ = _dataset_new


def _dict_new(cls, *args, **kwargs):
    # Get signature of __init__, map parameters and apply defaults. This
    # will raise an exception if there are any extra parameters.
    sig = inspect.Signature.from_callable(cls.__init__)
    sig = sig.replace(parameters=[v for k, v in sig.parameters.items() if k != "self"])
    mapping = sig.bind_partial(*args, **kwargs)
    mapping.apply_defaults()

    # The dictionary is now simply the arguments. Note that this means that
    # in contrast to the behaviour of AsDataset/AsDataarray, for
    # dictionaries we actually interpret a default of "None" as setting the
    # value in question to "None".
    instance = mapping.arguments

    # Check schema
    check.check_dict(instance, cls).expect()
    return instance


def dict_schema(cls):
    """Decorator for classes representing ``dict`` schemas, along the lines
    of :py:func:`xarray_dataarray_schema` and :py:func:`xarray_dataset_schema`.

    The annotated class can contain fields with arbitrary annotations, similar
    to a dataclass. They can be used with
    :py:func:`~xradio.schema.check.check_dict` for checking dictionieries
    against the schema. Furthermore, the class constructor will be overwritten
    to generate schema-confirming :py:class:`xarray.Dataset` objects.
    """

    # Make into a dataclass (might not even be needed at some point?)
    cls = dataclasses.dataclass(cls, init=True, repr=False, eq=False, frozen=True)

    # Make schema
    cls.__xradio_dict_schema = dataclass.xarray_dataclass_to_dict_schema(cls)

    # Replace __new__
    cls.__new__ = _dict_new

    return cls


def is_dict_schema(val: typing.Any):
    return type(val) == type and hasattr(val, "__xradio_dict_schema")


class AsDict:
    """Mix-in class to indicate dictionary data classes

    Deprecated - use decorator :py:func:`dict_schema` instead
    """

    __new__ = _dict_new



import os

import xarray as xr

from ._fits.xds_from_fits import _fits_image_to_xds


def _read_fits_image(
    infile: str, chunks: dict, verbose: bool, do_sky_coords: bool
) -> xr.Dataset:
    img_full_path = os.path.expanduser(infile)
    xds = _fits_image_to_xds(img_full_path, chunks, verbose, do_sky_coords)
    return xds



import astropy as ap
import astropy.units as u
import dask
import dask.array as da
import numpy as np
from typing import Dict, List
import xarray as xr
from xradio._utils.coord_math import _deg_to_rad
from xradio._utils.dict_helpers import make_quantity

_c = 2.99792458e08 * u.m / u.s
# OPTICAL = Z
_doppler_types = ["RADIO", "Z", "RATIO", "BETA", "GAMMA"]
_image_type = "image_type"


def _aperture_or_sky(xds: xr.Dataset) -> str:
    return "SKY" if "SKY" in xds.data_vars or "l" in xds.coords else "APERTURE"


def _get_xds_dim_order(has_sph: bool) -> list:
    dimorder = ["time", "polarization", "frequency"]
    dir_lin = ["l", "m"] if has_sph else ["u", "v"]
    dimorder.extend(dir_lin)
    return dimorder


def _convert_beam_to_rad(beam: dict) -> dict:
    """
    Convert something that looks like a CASA beam dictionary or close to
    to it to an xradio beam dict, with xradio quantities for major, minor,
    and pa, with the quantities converted to radians. Conversions are
    done using astropy. The input beam is not modified.
    """
    mybeam = {}
    for k in beam:
        try:
            q = u.quantity.Quantity(f"{beam[k]['value']}{beam[k]['unit']}")
        except:
            q = u.quantity.Quantity(f"{beam[k]['value']}{beam[k]['units']}")
        q = q.to("rad")
        j = "pa" if k == "positionangle" else k
        mybeam[j] = make_quantity(q.value, "rad")
    return mybeam


def _get_unit(u: str) -> str:
    if u == "'":
        return "arcmin"
    elif u == '"':
        return "arcsec"
    else:
        return u


def _coords_to_numpy(xds):
    for k, v in xds.coords.items():
        if dask.is_dask_collection(v):
            attrs = xds[k].attrs
            xds = xds.assign_coords({k: (v.sizes, v.to_numpy())})
            xds[k].attrs = attrs
    return xds


def _dask_arrayize_dv(xds: xr.Dataset) -> xr.Dataset:
    """
    If necessary, change data variables to dask arrays
    """
    for k, v in xds.data_vars.items():
        if not dask.is_dask_collection(v):
            dv_attrs = xds[k].attrs
            xds = xds.drop_vars([k])
            # may need to add sizes to this call as in numpy method analogs in this file
            xds = xds.assign({k: da.array(v)})
            xds[k].attrs = dv_attrs
    # only do the upper level data variables for now,
    # we don't have any data variables at sublevels so don't worry about them (yet)
    """
    if not is_copy:
        for k, v in xds.attrs.items():
            if isinstance(v, xr.Dataset):
                xds.attrs[k], is_copy = _dask_arrayize(v, is_copy)
    """
    return xds


def _numpy_arrayize_dv(xds: xr.Dataset) -> xr.Dataset:
    # just data variables right now
    # xds, is_copy = _coords_to_numpy(xds, is_copy)
    for k, v in xds.data_vars.items():
        if dask.is_dask_collection(v):
            attrs_dv = xds[k].attrs
            xds = xds.drop_vars([k])
            xds = xds.assign({k: (v.sizes, v.to_numpy())})
            xds[k].attrs = attrs_dv
    """
    for k, v in xds.attrs.items():
        if isinstance(v, xr.Dataset):
            xds.attrs[k], is_copy = _dask_arrayize(v, is_copy)
    """
    return xds


def _default_freq_info() -> dict:
    return {
        "rest_frequency": make_quantity(1420405751.7860003, "Hz"),
        "type": "frequency",
        "frame": "LSRK",
        "units": "Hz",
        "waveUnit": "mm",
        "cdelt": 1000.0,
        "crval": 1415000000.0,
    }


def _freq_from_vel(
    crval: float,
    cdelt: float,
    crpix: float,
    cunit: str,
    ctype: str,
    nchan: float,
    restfreq: float,
) -> tuple:
    """
    inputs are velocity info
    restfreq is an astropy quantity with frequency units
    a tuple of two dicts is returned. The first dict represents
    the frequencies, the second the velocities. Both dicts have
    keys 'value', 'unit', 'crval', 'cdelt', 'crpix'. 'value'
    is a list.
    """
    v0 = crval - cdelt * crpix
    vel = [v0 + i * cdelt for i in range(nchan)]
    vel = vel * u.Unit(cunit)
    v_dict = {
        "value": vel.value,
        "unit": cunit,
        "crval": crval,
        "cdelt": cdelt,
        "crpix": crpix,
    }
    uctype = ctype.lower()
    if uctype in ["z", "optical"]:
        freq = restfreq / (np.array(vel.value) * vel.unit / _c + 1)
        freq = freq.to(u.Hz)
        fcrval = restfreq / (crval * vel.unit / _c + 1)
        fcdelt = -restfreq / _c / (crval * vel.unit / _c + 1) ** 2 * cdelt * vel.unit
        f_dict = {
            "value": freq.value,
            "unit": "Hz",
            "crval": fcrval.to(u.Hz).value,
            "cdelt": fcdelt.to(u.Hz).value,
            "crpix": crpix,
        }
    else:
        raise RuntimeError(f"Unhandled doppler type {ctype}")
    return f_dict, v_dict


def _compute_world_sph_dims(
    projection: str,
    shape: List[int],  # two element list of long-lat shape
    ctype: List[str],  # two element list of long-lat axis names
    crpix: List[float],  # two element list of long-lat crpix (zero-based)
    crval: List[float],  # two element list of long-lat crval
    cdelt: List[float],  # two element list of long-lat increments
    cunit: List[str],  # two element list of long-lat units
) -> dict:
    # Note that if doesn't matter if the inputs are in long, lat or lat, long order,
    # as long as all inputs have consistent ordering
    wcs_dict = {}
    ret = {
        "axis_name": [None, None],
        "ref_val": [None, None],
        "inc": [None, None],
        "unit": ["rad", "rad"],
        "value": [None, None],
    }
    for i in range(2):
        axis_name = ctype[i].lower()
        if axis_name.startswith("right") or axis_name.startswith("ra"):
            fi = 1
            wcs_dict[f"CTYPE1"] = f"RA---{projection}"
            new_name = "right_ascension"
        if axis_name.startswith("dec"):
            fi = 2
            wcs_dict["CTYPE2"] = f"DEC--{projection}"
            new_name = "declination"
        wcs_dict[f"NAXIS{fi}"] = shape[i]
        j = fi - 1
        x_unit = _get_unit(cunit[i])
        wcs_dict[f"CUNIT{fi}"] = x_unit
        wcs_dict[f"CDELT{fi}"] = cdelt[i]
        # FITS arrays are 1-based
        wcs_dict[f"CRPIX{fi}"] = crpix[i] + 1
        wcs_dict[f"CRVAL{fi}"] = crval[i]
        ret["axis_name"][j] = new_name
        ret["ref_val"][j] = u.quantity.Quantity(f"{crval[i]}{x_unit}").to("rad").value
        ret["inc"][j] = u.quantity.Quantity(f"{cdelt[i]}{x_unit}").to("rad").value
    w = ap.wcs.WCS(wcs_dict)
    x, y = np.indices(w.pixel_shape)
    long, lat = w.pixel_to_world_values(x, y)
    # long, lat from above eqn will always be in degrees, so convert to rad
    ret["value"][0] = long * _deg_to_rad
    ret["value"][1] = lat * _deg_to_rad
    return ret


def _compute_velocity_values(
    restfreq: float,  # in Hz
    freq_values: List[float],  # in Hz
    doppler: str,  # doppler definition
) -> List[float]:
    dop = doppler.lower()
    if dop == "radio":
        return [((1 - f / restfreq) * _c).value for f in freq_values]
    elif dop in ["z", "optical"]:
        return [((restfreq / f - 1) * _c).value for f in freq_values]
    else:
        raise RuntimeError(f"Doppler definition {doppler} not supported")


def _compute_linear_world_values(
    naxis: int, crval: float, crpix: float, cdelt: float
) -> np.ndarray:
    """
    Simple linear transformation to get world values which can be used for certain
    coordinates like (often) frequency, l, m, u, v
    """
    return np.array([crval + (i - crpix) * cdelt for i in range(naxis)])


# def _compute_ref_pix(xds: xr.Dataset, direction: dict) -> np.ndarray:
def _compute_sky_reference_pixel(xds: xr.Dataset) -> np.ndarray:
    crpix = [None, None]
    for i, c in enumerate(["l", "m"]):
        x = xds[c].values
        y = np.array(range(len(x)))
        if x[1] < x[0]:
            x = x[::-1]
            y = y[::-1]
        crpix[i] = np.interp(0.0, x, y)
    return np.array(crpix)


def _l_m_attr_notes() -> Dict[str, str]:
    return {
        "l": "l is the angle measured from the phase center to the east. "
        "So l = x*cdelt, where x is the number of pixels from the phase center. "
        "See AIPS Memo #27, Section III.",
        "m": "m is the angle measured from the phase center to the north. "
        "So m = y*cdelt, where y is the number of pixels from the phase center. "
        "See AIPS Memo #27, Section III.",
    }


def _set_multibeam_array(xds, beam_ary, units):
    if "beam" in xds.attrs:
        if xds.attrs["beam"] is None:
            del xds.attrs["beam"]
        else:
            raise RuntimeError(
                "Error: xds has beam attr. It must be removed "
                "before multiple beams can be added"
            )
    xdb = xr.DataArray(
        beam_ary, dims=["time", "polarization", "frequency", "beam_param"]
    )
    xdb = xdb.rename("beam")
    xdb = xdb.assign_coords(beam_param=["major", "minor", "pa"])
    xdb.attrs["units"] = units
    xds["beam"] = xdb
    return xds
